import copy
import math
import os
import numpy as np
from astropy.io import fits
from . import cosutil
from . import dispersion
from . import extract
from . import shiftfile
from . import wavecal
from . import ccos
from .calcosparam import *       # parameter definitions
from . import findshift1

# add DELTA_SHIFT[12] to a shift for segment A to get the shift for B
DELTA_SHIFT1 = 0.0
DELTA_SHIFT2 = 0.0

# This is used in makeHistogram.  If the maximum value of the wavecal
# count rate within the search region for a lamp flash is less than or
# equal to this value, it will be assumed either that the lamp was not on
# or that the search region did not cover the wavecal spectrum.
MIN_COUNT_RATE = 20.

# This is the nominal location of an NUV wavecal image; X0 is in the more
# rapidly varying axis, and Y0 is in the less rapidly varying axis.
# xxx these should be gotten from a reference table
X0 = 600.
Y0 = 606.
# search range in each axis
DX = 50
DY = 50

def processConcurrentWavecal(events, outflash, shift_file,
                             info, switches, reffiles, phdr, hdr):
    """Determine shifts from concurrent (tagflash) wavecal exposures.

    If tagflash mode was not used or the wavecorr switch is "OMIT" or
    "SKIPPED", this function will return without doing anything.

    Parameters
    ----------
    events: array_like
        Data block for a corrtag table.

    outflash: str
        Name of output file for extracted wavecal spectra.

    shift_file: str or None
        If not None, this text file contains values of shift1 (and
        possibly shift2) to override the values found via wavecal
        processing.

    info: dictionary
        Header keywords and values.

    switches: dictionary
        Calibration switch keywords and values.

    reffiles: dictionary
        Reference file names.

    phdr: FITS Header object
        Primary header of corrtag file.

    hdr: FITS Header object
        Events extension header of corrtag file.

    Returns
    -------
    (tl_time, shift1_vs_time, wavecorr): tuple of three items
        tl_time is the array of times at one-second intervals, for the
        timeline table.  shift1_vs_time is the array of corresponding
        values of shift1a or shift1b, or None if there were no flashes or
        if the exposure time is zero.  wavecorr will normally be "COMPLETE"
        but may be some other value, e.g. "SKIPPED", if wavecal processing
        was not actually done.
    """

    tl_time = np.zeros(1, dtype=np.float32)     # replaced later

    if not info["tagflash"]:
        return (tl_time, None, switches["wavecorr"])

    # This test allows processing to continue if wavecorr is either
    # PERFORM or COMPLETE.
    if switches["wavecorr"] == "OMIT" or switches["wavecorr"] == "SKIPPED":
        cw = initWavecal(events, outflash, shift_file,
                         info, switches, reffiles, phdr, hdr)
        tl_time = cosutil.timelineTimes(self.time[0], self.time[-1], dt=1.)
        return (tl_time, None, switches["wavecorr"])

    cosutil.printMsg("Process tagflash wavecal")
    wavecal.printWavecalRef(reffiles)
    cosutil.printRef("disptab", reffiles)

    cw = initWavecal(events, outflash, shift_file, info, switches, reffiles,
                     phdr, hdr)
    if cw.override_segment_B:
        if cw.segment_A_present:
            # Copy lampflash_a.fits to lampflash_b.fits.
            cw.copySegAtoSegB()
            cw.applyCorrections()
            (avg_dx, avg_dy) = cw.avgShift()
            cw.setShiftKeywords(avg_dx, avg_dy)
            (tl_time, shift1_vs_time) = cw.shift1VsTime()
            cw.closeOutFlash()
        else:
            cw.setSegBtoZero()          # set shifts to fp_pixel_shift
            cw.setUpOutFlash()
            (avg_dx, avg_dy) = cw.miscSegB()
            cw.applyCorrections()
            cw.setShiftKeywords(avg_dx, avg_dy)
            (tl_time, shift1_vs_time) = cw.shift1VsTime()
            cw.writeOutFlash()
        phdr["wavecorr"] = cw.wavecorr
        return (tl_time, shift1_vs_time, cw.wavecorr)

    cw.getStartStopTimes()
    if cw.numflash < 1 and cw.shift_file is not None:
        cw.dummyFlash()
    if cw.numflash < 1:
        # write an empty lampflash table
        cosutil.printWarning("No lamp flash was found.")
        cw.wavecorr = "SKIPPED"
        cw.setUpOutFlash()
        phdr["wavecorr"] = cw.wavecorr
        phdr["lampused"] = "NONE"
        cw.ofd[0].header["lampused"] = "NONE"
        cw.writeOutFlash()
        return (tl_time, None, cw.wavecorr)

    cw.setUpOutFlash()          # create an HDU list for the outflash file

    cw.getWavecalParameters()
    cw.findShifts()
    cw.applyCorrections()

    (avg_dx, avg_dy) = cw.avgShift()
    cw.setShiftKeywords(avg_dx, avg_dy)

    (tl_time, shift1_vs_time) = cw.shift1VsTime()

    if switches["statflag"] == "PERFORM":
        cw.doStat()
    phdr["wavecorr"] = cw.wavecorr

    cw.writeOutFlash()

    return (tl_time, shift1_vs_time, cw.wavecorr)

def r_key(x):
    """Comparison key for sorting locations in setRegions().

    The comparison is based entirely on the second element, b_spec.
    """

    return x[1]

def initWavecal(events, outflash, shift_file, info, switches, reffiles,
                phdr, hdr):
    """Return a ConcurrentWavecal object, depending on detector.

    Parameters
    ----------
    events: array_like
        Data block for an events table corrtag file.

    outflash: str
        Name of output file for table of extracted wavecal spectra.

    shift_file: str or None
        Name of user-supplied file to override shifts (or None).

    info: dictionary
        Header keywords and values.

    switches: dictionary
        Calibration switches.

    reffiles: dictionary
        Reference file names.

    phdr: FITS Header object
        Primary header of corrtag file.

    hdr: FITS Header object
        Events extension header.
    """

    if info["detector"] == "FUV":
        cw = FUVConcurrentWavecal(events, outflash, shift_file,
                                  info, switches, reffiles, phdr, hdr)
    elif info["obstype"] == "IMAGING":
        cw = NUVImagingWavecal(events, outflash, shift_file,
                               info, switches, reffiles, phdr, hdr)
    else:
        cw = NUVConcurrentWavecal(events, outflash, shift_file,
                                  info, switches, reffiles, phdr, hdr)

    return cw

class ConcurrentWavecal(object):
    """Process wavecals embedded in a science observation (tagflash).

    @ivar events: data block for a corrtag table
    @type events: record array

    @ivar outflash: name of output file for extracted wavecal spectra
    @type outflash: string

    @ivar shift_file: name of user-supplied file to override shifts
    @type shift_file: string

    @ivar info: dictionary of header keywords and values
    @type info: dictionary

    @ivar switches: calibration switches
    @type switches: dictionary

    @ivar reffiles: dictionary of reference file names
    @type reffiles: dictionary

    @ivar phdr: primary header of corrtag file
    @type phdr: FITS header object

    @ivar hdr: events extension header of corrtag file
    @type hdr: FITS header object
    """

    def __init__(self, events, outflash, shift_file,
                 info, switches, reffiles, phdr, hdr,
                 delta_t=0.2, buffer_on=2.0, buffer_off=4.0):

        self.events = events
        self.outflash = outflash
        self.shift_file = shift_file
        self.info = info
        self.switches = switches
        self.reffiles = reffiles
        self.phdr = phdr
        self.hdr = hdr
        self.wavecorr = "COMPLETE"      # possibly reset later
        self.user_shifts = None
        self.override_segment_B = False
        self.segment_A_present = True

        self.ofd = None                 # HDU list for outflash FITS file

        # lamp_on and lamp_off are in seconds, with the same zero point
        # as the TIME column in the events table
        self.lamp_on = []               # start time of each flash
        self.lamp_off = []              # stop time of each flash
        self.lamp_duration = []         # length of each flash
        self.lamp_median = []           # median time of each flash
        self.numflash = 0               # number of flashes
        self.lamp_is_on = False         # true if any flash was actually on

        # Did the user supply a file with overrides for the shifts?
        if self.shift_file is not None:
            self.user_shifts = shiftfile.ShiftFile(self.shift_file,
                                                   self.info["root"],
                                                   self.info["fpoffset"])
        else:
            self.user_shifts = None

        # will be the count rate in each delta_t time interval in the
        # wavecal region of the detector
        self.src_counts = None
        self.delta_t = delta_t          # bin events in this size interval
        # subtract buffer_on from lamp_on and add buffer_off to lamp_off
        # before searching for actual lamp turn-on and turn-off times
        self.buffer_on = buffer_on
        self.buffer_off = buffer_off

        # for FUV segment_list is just one segment name;
        # for NUV this is all three stripe names ["NUVA", "NUVB", "NUVC"].
        self.segment_list = []          # segment names

        # the key is segment or stripe name, the value is a list, each
        # element of which is a two-element list of the slice limits
        # (in the cross-dispersion direction) within which the shifts
        # should be subtracted from the pixel coordinates
        self.regions = {}               # apply shift1 only to these regions

        # the number of elements is the number of flashes:
        # each element is a dictionary; the key is segment or stripe name,
        # and the value is the shift in the dispersion direction
        self.shift1 = []                # dictionaries of shift in disp dir
        self.chi_square = []            # dictionaries of Chi square
        self.n_deg_freedom = []         # dictionaries of num of deg of freedom

        # each element is the shift in the cross-dispersion direction
        # (one value for all stripes)
        self.shift2 = []                # shifts in cross-dispersion direction

        # the relevant row from the wavecal parameters table
        self.wcp_info = None            # matching row (just one) from wcp table

        # times of photon events, in seconds since EXPSTART
        self.time = events.field("TIME").astype(np.float64)

        # These five columns are assigned by a subclass, depending on
        # the detector.  xi is in the dispersion direction, and eta is
        # in the cross-dispersion direction.
        self.xi = None
        self.eta = None
        self.dq = None                  # data quality flags; 0 is OK
        self.xi_corr = None             # corrected coords (XFULL)
        self.eta_corr = None            # corrected coords (YFULL)
        self.spectrum = None            # scratch space for extracted spectrum

    def dummyFlash(self):
        """Assign values for one flash, even if there wasn't."""

        self.numflash = 1
        self.lamp_on = [0.]
        self.lamp_off = [1.]
        self.lamp_duration = [1.]
        self.lamp_median = [0.5]

    def setUpOutFlash(self):
        """Create an HDU list for the outflash FITS table."""

        if not self.outflash:
            self.ofd = None
            return

        # Number of elements in a WAVELENGTH or GROSS array.
        rpt = str(len(self.spectrum))

        col = []
        col.append(fits.Column(name="SEGMENT", format="4A"))
        col.append(fits.Column(name="TIME", format="1D",
                               disp="F8.3", unit="s"))
        col.append(fits.Column(name="EXPTIME", format="1D",
                               disp="F8.3", unit="s"))
        col.append(fits.Column(name="LAMP_ON", format="1D",
                               disp="F8.3", unit="s"))
        col.append(fits.Column(name="LAMP_OFF", format="1D",
                               disp="F8.3", unit="s"))
        col.append(fits.Column(name="NELEM", format="1J",
                               disp="I6"))
        col.append(fits.Column(name="WAVELENGTH", format=rpt+"D",
                               unit="angstrom"))
        col.append(fits.Column(name="NET", format=rpt+"E",
                               unit="count /s"))
        col.append(fits.Column(name="GROSS", format=rpt+"E",
                               unit="count /s"))
        col.append(fits.Column(name="BACKGROUND", format=rpt+"E",
                               unit="count /s"))
        col.append(fits.Column(name="SHIFT_DISP", format="1E",
                               unit="pixel"))
        col.append(fits.Column(name="SHIFT_XDISP", format="1E",
                               unit="pixel"))
        col.append(fits.Column(name="SPEC_FOUND", format="1L"))
        col.append(fits.Column(name="CHI_SQUARE", format="1E"))
        col.append(fits.Column(name="N_DEG_FREEDOM", format="1J",
                               disp="I5"))
        cd = fits.ColDefs(col)

        nrows = self.numflash * len(self.segment_list)

        primary_hdu = fits.PrimaryHDU(header=self.phdr)
        self.ofd = fits.HDUList(primary_hdu)

        #
        # The WCS table keywords don't need to be propagated to the lampflash
        # extension, so we can remove them before we create the hdu
        newheader = self.hdr.copy()
        self.deleteCoordinateKeywords(newheader)

        hdu = fits.BinTableHDU.from_columns(cd, header=newheader, nrows=nrows)
        hdu.name = "LAMPFLASH"
        self.ofd.append(hdu)

        cosutil.updateFilename(self.ofd[0].header, self.outflash)
        nextend = len(self.ofd) - 1
        self.ofd[0].header["nextend"] = nextend
        self.ofd[0].header["wavecorr"] = self.wavecorr

        # We know this value, so assign it now.
        self.ofd[1].data.field("nelem")[:] = len(self.spectrum)

    def deleteCoordinateKeywords(self, header):
        """Delete keywords that are not relevant for extracted spectra.

        Parameters
        ----------
        header: FITS header object
            Header for table of extracted wavecal spectra (will be modified
            in-place).
        """

        ikey = ["TCTYP2", "TCTYP3", "TCRVL2", "TCRVL3", "TCRPX2", "TCRPX3",
                "TCDLT2", "TCDLT3", "TCUNI2", "TCUNI3",
                "TC2_2",  "TC2_3",  "TC3_2",  "TC3_3",
                "TALEN2", "TALEN3"]
        for keyword in ikey:
            if keyword in header:
                del header[keyword]

        # Set the values of these keywords to zero.
        zkey = ["SHIFT1A", "SHIFT1B", "SHIFT1C"]
        for keyword in zkey:
            if keyword in header:
                header[keyword] = 0.

    def doStat(self):
        """Compute mean and max of the GROSS column."""

        if self.ofd is not None:
            cosutil.doTagFlashStat(self.ofd)

    def writeOutFlash(self):
        """Write the outflash HDU list to the output file."""

        if self.ofd is not None:
            self.ofd.writeto(self.outflash, output_verify="fix")
            self.ofd.close()

    def copyColumns(self):
        """Copy xi and eta to xi_corr and eta_corr."""

        self.xi_corr[:] = self.xi.copy()
        self.eta_corr[:] = self.eta.copy()

    def getStartStopTimes(self):
        """Get the times when the lamp was turned on or off."""

        # total number of tagflash wavecal exposures (commanded value)
        self.numflash = self.hdr.get("NUMFLASH", default=0)

        t0 = self.time[0]
        t_last = self.time[-1]
        numflash = 0            # will be actual number of tagflash wavecals

        # First assign reasonable initial values, based on header keywords.
        if self.info["tagflash_type"] == TAGFLASH_TYPE_AUTO:
            # Explicitly get LMP_ONi and LMPOFFi for every flash.
            for n in range(self.numflash):
                key_on = "LMP_ON%d" % (n+1)     # one indexed for FITS
                key_off = "LMPOFF%d" % (n+1)
                key_duration = "LMPDUR%d" % (n+1)
                starttime = self.hdr.get(key_on)
                endtime = self.hdr.get(key_off)
                duration = self.hdr.get(key_duration)
                if endtime < t0 or starttime > t_last:
                    continue
                starttime = max(starttime, t0)
                endtime = min(endtime, t_last)
                duration = min(duration, endtime-starttime)
                self.lamp_on.append(starttime)
                self.lamp_off.append(endtime)
                self.lamp_duration.append(duration)
                self.lamp_median.append((starttime + endtime) / 2.)
                numflash += 1
        else:
            # The flashes are uniformly spaced, so calculate the
            # expected on and off times.
            starttime = self.hdr.get("LMP_ON1")
            starttime = max(starttime, t0)
            endtime = self.hdr.get("LMPOFF1")
            duration = self.hdr.get("LMPDUR1")
            if self.numflash > 1:
                spacing = self.hdr.get("LMP_ON2") - self.hdr.get("LMP_ON1")
            else:
                spacing = 1.
            for n in range(self.numflash):
                endtime = starttime + duration
                if starttime > t_last:
                    break
                self.lamp_on.append(starttime)
                self.lamp_off.append(endtime)
                self.lamp_duration.append(duration)
                self.lamp_median.append((starttime + endtime) / 2.)
                starttime += spacing
                numflash += 1

        self.numflash = numflash        # can be modified by findFlash

        # Determine the actual times from the data, and then set the keywords
        # and update the attributes.
        self.findFlash(output=None, update=True)

    def getWavecalParameters(self):
        """Get the matching row from the wavecal parameters table."""

        if self.info["obstype"] == "IMAGING":
            self.wcp_info = None
        else:
            wcp_info = cosutil.getTable(self.reffiles["wcptab"],
                            filter={"opt_elem": self.info["opt_elem"]},
                            exactly_one=True)
            self.wcp_info = wcp_info[0]

    def findShifts(self):
        """For each wavecal flash, find the shift in each axis."""

        xtractab = self.reffiles["xtractab"]
        disptab = self.reffiles["disptab"]
        lamptab = self.reffiles["lamptab"]
        # segment will be added to the filter within the loop.  Note that
        # the aperture is explicitly set to "WCA", because the aperture
        # keyword will give the aperture used for the science data.
        filter_1dx = {"opt_elem": self.info["opt_elem"],
                      "cenwave": self.info["cenwave"],
                      "aperture": "WCA"}
        filter_disp = {"opt_elem": self.info["opt_elem"],
                       "cenwave": self.info["cenwave"],
                       "aperture": "WCA",
                       "fpoffset": self.info["fpoffset"]}
        filter_lamp = {"opt_elem": self.info["opt_elem"],
                       "cenwave": self.info["cenwave"]}
        # These two flags will be used within a loop below.
        got_fpoffset = cosutil.findColumn(lamptab, "fpoffset")
        got_pixel_shift = cosutil.findColumn(lamptab, "fp_pixel_shift")
        if got_fpoffset:
            filter_lamp["fpoffset"] = self.info["fpoffset"]
        xc_range = self.wcp_info.field("xc_range")
        stepsize = self.wcp_info.field("stepsize")
        xd_range = self.wcp_info.field("xd_range")
        box = self.wcp_info.field("box")
        try:
            search_offset = self.wcp_info.field("search_offset")
        except KeyError:
            search_offset = 0.
        # initial_offset is used as the center of the search range, when
        # matching the spectrum to the template.
        if got_pixel_shift:
            initial_offset = search_offset
        else:
            initial_offset = self.info["fpoffset"] * stepsize + search_offset

        # Create a data quality array, and assign values from the bpixtab.
        if self.info["detector"] == "FUV":
            axis_height = FUV_Y
            axis_length = FUV_EXTENDED_X
        else:
            axis_height = NUV_Y
            axis_length = NUV_EXTENDED_X
        # create and populate a DQ array
        dq_array = np.zeros((axis_height,axis_length), dtype=np.int16)
        cosutil.updateDQArray(self.info, self.reffiles, dq_array,
                              {(0, 1024): [0., 0., 0., 0.]},
                              (0., 0.), -10, None)
        # weights from flat field or nonlinearity
        epsilon = self.events.field("epsilon")

        # Find the offsets in both axes, for each wavecal exposure.
        row = 0         # incremented in the second loop over segments
        cosutil.printMsg(
"  segment    cross-disp           dispersion direction", VERBOSE)
        cosutil.printMsg(
"            shift (locn)      shift err  [orig.]    FP   chi sq (n)", VERBOSE)
        cosutil.printMsg(
"  -------   -------------     -------------------------  ----------", VERBOSE)
        for n in range(self.numflash):
            (i0, i1) = ccos.range(self.time, self.lamp_on[n], self.lamp_off[n])

            # Find offset from nominal in cross-dispersion direction.
            (shift2, xd_shifts, xd_locn, lamp_is_on) = \
                wavecal.ttFindWavecalSpectrum(
                        self.xi[i0:i1], self.eta[i0:i1], self.dq[i0:i1],
                        self.info, xd_range, box, xtractab)
            if shift2 is None:
                shift2 = 0.
            self.shift2.append(shift2)
            if not self.lamp_is_on:
                self.lamp_is_on = lamp_is_on

            # Extract wavecal spectra from events table, and determine offset
            # in dispersion direction.
            x_offset = self.info["x_offset"]    # offset of lamptab in template
            sdqflags = self.info["sdqflags"]
            shift1 = {}                 # to be saved in an attribute
            chi_square = {}             # to be saved in an attribute
            n_deg_freedom = {}          # to be saved in an attribute
            save_spectra = {}   # save, but just within this loop
            save_net = {}       # save net counts, within this loop
            save_bkg = {}       # save background counts, within this loop
            save_templates = {}
            fp_pixel_shift = {}
            spec_found = {}     # true if spectrum was found
            lamp_found = {}     # for second loop over segments
            at_least_one_found = False
            ((user_shift1, user_shift2), nfound) = ((None, None), 0)
            for segment in self.segment_list:   # first loop over segments
                # Check the user-supplied shift file for a match with this
                # flash number and segment/stripe; we don't need the value
                # yet, but if there's a match we want to set the flag to
                # say that the shift was found.
                if self.user_shifts is not None:
                    ((user_shift1, user_shift2), nfound) = \
                        self.user_shifts.getShifts((n+1, segment))
                spec_found[segment] = False     # initial value
                filter_1dx["segment"] = segment
                filter_lamp["segment"] = segment
                xtract_info = cosutil.getTable(xtractab, filter_1dx,
                                               at_least_one=True)
                snr_ff = 0.             # ignore error array
                axis = 1                # dispersion is along X axis
                dummy_exptime = 1.
                (N_i, ERR_i, ERR_LOW_i, VARIANCE_FLAT_i, VARIANCE_COUNTS_i, VARIANCE_BKG_i,
            GC_i, GCOUNTS_i, BK_i, DQ_i, DQ_WGT_i,
            DQ_ALL_i, LOWER_OUTER_INDEX_i, UPPER_OUTER_INDEX_i,
            LOWER_INNER_INDEX_i, UPPER_INNER_INDEX_i,
            ENCLOSED_FRACTION_i, AV_E_BKG_i,
            LOWER_OUTER_VALUE_i, LOWER_INNER_VALUE_i,
            UPPER_INNER_VALUE_i, UPPER_OUTER_VALUE_i) = \
                    extract.extractCorrtag(self.xi[i0:i1], self.eta[i0:i1],
                                self.dq[i0:i1], epsilon[i0:i1], dq_array,
                                self.ofd[1].header, segment, axis_length,
                                x_offset, sdqflags, snr_ff,
                                dummy_exptime, self.switches["backcorr"],
                                axis, xtract_info, 0., shift2)
                self.spectrum = GCOUNTS_i       # gross counts
                save_spectra[segment] = self.spectrum.copy()
                save_net[segment] = N_i         # net counts
                save_bkg[segment] = BK_i        # background counts
                lamp_found[segment] = True      # default
                lamp_info = cosutil.getTable(lamptab, filter_lamp,
                                             at_least_one=True)
                raw_template = lamp_info.field("intensity")[0]
                save_templates[segment] = cosutil.getTemplate(raw_template,
                                          x_offset, len(self.spectrum))
                if got_pixel_shift:
                    fp_pixel_shift[segment] = \
                            lamp_info.field("fp_pixel_shift")[0]
                else:
                    fp_pixel_shift[segment] = 0.
                if user_shift1 is not None or xd_shifts[segment] is not None:
                    spec_found[segment] = True
                    at_least_one_found = True

            if not at_least_one_found:
                # flag all segments/stripes as not found
                for segment in self.segment_list:
                    spec_found[segment] = False

            # find offset in dispersion direction
            fs1 = findshift1.Shift1(save_spectra, save_templates,
                                    self.info, self.reffiles,
                                    xc_range, fp_pixel_shift, initial_offset,
                                    spec_found)
            fs1.findShifts()

            # Print results, and save extracted spectra in lampflash table.
            for segment in self.segment_list:   # second loop over segments
                user_specified = False
                if self.user_shifts is not None:        # override shifts?
                    # note that flash number is one indexed
                    ((user_shift1, user_shift2), nfound) = \
                        self.user_shifts.getShifts((n+1, segment))
                    if user_shift1 is not None:
                        fs1.setShift1(segment, user_shift1)
                        user_specified = True
                shift1[segment] = fs1.getShift1(segment)
                measured_shift1 = fs1.getMeasuredShift1(segment)
                fp_pixel_shift_seg = fs1.getFpPixelShift(segment)
                chi_square[segment] = fs1.getChiSq(segment)
                n_deg_freedom[segment] = fs1.getNdf(segment)
                filter_disp["segment"] = segment
                if not lamp_found[segment]:
                    # no matching row in table
                    foundit = False
                    message = \
                        "%2d %4s skipped due to missing reference table row" \
                        % (n+1, segment)
                else:
                    foundit = fs1.getSpecFound(segment)
                    if xd_shifts[segment] is None:
                        # ttFindWavecalSpectrum couldn't find the spectrum
                        message = \
"%2d %4s      ---- (%5.1f) %9.1f %4.2f [%5.1f]        %6.1f (%d)  # not found in XD" \
                            % (n+1, segment,
                               xd_locn[segment], shift1[segment],
                               fs1.getScatter(segment), measured_shift1,
                               chi_square[segment], n_deg_freedom[segment])
                    else:
                        message = \
"%2d %4s %9.1f (%5.1f) %9.1f %4.2f [%5.1f] %6.1f  %6.1f (%d)" \
                            % (n+1, segment, xd_shifts[segment],
                               xd_locn[segment], shift1[segment],
                               fs1.getScatter(segment), measured_shift1,
                               fp_pixel_shift_seg,
                               chi_square[segment], n_deg_freedom[segment])
                        if not foundit:
                            message = message + "  # not found"
                    if user_specified:
                        message = message + "  # user-specified"
                cosutil.printMsg(message, VERBOSE)
                # copy to outflash table data
                if lamp_found[segment]:
                    self.saveSpectrum(disptab, filter_disp, n, row,
                                      save_spectra[segment],
                                      save_net[segment], save_bkg[segment],
                                      shift1[segment], xd_shifts[segment],
                                      foundit, chi_square[segment],
                                      n_deg_freedom[segment])
                else:
                    self.saveSpectrum(disptab, filter_disp,
                                      n, row, None, None, None, None, None,
                                      foundit, chi_square[segment],
                                      n_deg_freedom[segment])
                row += 1
            if self.info["detector"] == "NUV":
                cosutil.printMsg("%2d      avg %5.1f" % (n+1, shift2), VERBOSE)
            self.shift1.append(shift1)
            self.chi_square.append(chi_square)
            self.n_deg_freedom.append(n_deg_freedom)

    def saveSpectrum(self, disptab, filter, n, row,
                     spectrum, net_spectrum, bkg_spectrum,
                     shift1, shift2, spec_found,
                     chi_square, n_deg_freedom):
        """Copy the spectrum to the record array for the outflash table.

        Parameters
        ----------
        disptab: str
            Name of the dispersion relation table.

        filter: dictionary
            For extracting the row from the disptab.

        n: int
            Index of current lamp flash.

        row: int
            Row index (zero indexed) in output table.

        spectrum: array_like or None
            Gross counts for current segment or stripe (may be None).

        net_spectrum: array_like or None
            Net counts for current segment or stripe (may be None).

        bkg_spectrum: array_like or None
            Background counts for current segment or stripe (may be None).

        shift1: float
            Shift in dispersion direction.

        shift2: float or None
            Shift in cross-dispersion direction (may be None).

        spec_found: boolean
            Was the wavecal spectrum actually found?

        chi_square: float
            Chi square for the current flash.

        n_deg_freedom: int
            Number of degrees of freedom for current flash.
        """

        if self.ofd is None:
            return

        t0 = self.lamp_on[n]
        t1 = self.lamp_off[n]

        if spectrum is not None:
            pixel = np.arange(len(spectrum), dtype=np.float64)
            disp_rel = dispersion.Dispersion(disptab, filter)
            pixel -= shift1         # correct the wavelengths for the shift
            # Correct for any extra pixels in the dispersion direction.
            pixel -= self.info["x_offset"]
            wavelength = disp_rel.evalDisp(pixel)
            disp_rel.close()

        self.ofd[1].data.field("segment")[row] = filter["segment"]
        self.ofd[1].data.field("time")[row] = self.lamp_median[n]
        self.ofd[1].data.field("exptime")[row] = t1 - t0
        self.ofd[1].data.field("lamp_on")[row] = self.lamp_on[n]
        self.ofd[1].data.field("lamp_off")[row] = self.lamp_off[n]
        if spectrum is None:
            self.ofd[1].data.field("wavelength")[row][:] = 0.
        else:
            self.ofd[1].data.field("wavelength")[row] = wavelength
        exptime = t1 - t0
        if exptime <= 0.:
            exptime = 1.
        if spectrum is None:
            self.ofd[1].data.field("gross")[row][:] = 0.
            self.ofd[1].data.field("net")[row][:] = 0.
            self.ofd[1].data.field("background")[row][:] = 0.
            self.ofd[1].data.field("shift_disp")[row] = 0.
        else:
            self.ofd[1].data.field("gross")[row] = spectrum / exptime
            self.ofd[1].data.field("net")[row] = net_spectrum / exptime
            self.ofd[1].data.field("background")[row] = bkg_spectrum / exptime
            self.ofd[1].data.field("shift_disp")[row] = shift1
        if shift2 is None:
            self.ofd[1].data.field("shift_xdisp")[row] = 0.
        else:
            self.ofd[1].data.field("shift_xdisp")[row] = shift2
        self.ofd[1].data.field("spec_found")[row] = spec_found
        self.ofd[1].data.field("chi_square")[row] = chi_square
        self.ofd[1].data.field("n_deg_freedom")[row] = n_deg_freedom

    def applyCorrections(self):
        """Apply the shift1[a-c] and shift2[a-c] offsets."""

        # There will be at least one flash.
        for n in range(self.numflash):
            (i0, i1, extrapolate) = self.getInterval(n)
            self.shift2Corr(n, i0, i1, extrapolate)
            self.shift1Corr(n, i0, i1, extrapolate)

    def getInterval(self, n):
        """Get the slice for times between wavecal exposures n and n+1.

        Parameters
        ----------
        n: int
            Index of flash at left end of time interval; the valid range
            for n is from 0 to the number of flashes minus one, inclusive.

        Returns
        -------
        (i0, i1, extrapolate): tuple of int, int, boolean
            i0 and i1 are the indices in the time array (Python slice
            indices) over which to apply the shift based on flashes n and
            n+1 (unless n is the last), and extrapolate is True if
            n is the last flash.
        """

        if n == 0:
            t0 = self.time[0]           # extrapolate to the beginning
        else:
            t0 = self.lamp_median[n]

        if n == self.numflash - 1:
            extrapolate = True
            t1 = self.time[-1]
        else:
            extrapolate = False
            t1 = self.lamp_median[n+1]

        (i0, i1) = ccos.range(self.time, t0, t1)
        if extrapolate:
            i1 = len(self.time)

        return (i0, i1, extrapolate)

    def shift1Corr(self, n, i0, i1, extrapolate=False):
        """Correct the pixel coordinates in the dispersion direction.

        Parameters
        ----------
        n: int
            Apply shift1 for the nth time interval between wavecals.

        i0, i1: int
            [i0:i1] is the slice of event numbers to be corrected.

        extrapolate: boolean
            True if n is the last flash.
        """

        for segment in self.segment_list:

            # Restrict the correction to the applicable regions.
            shift_flags = np.zeros(i1 - i0, dtype=np.bool_)
            locn_list = self.regions[segment]
            for region in locn_list:
                if region[0] is None:
                    shift_flags |= np.where(
                                   self.eta[i0:i1] < region[1], True, False)
                elif region[1] is None:
                    shift_flags |= np.where(
                                   self.eta[i0:i1] >= region[0], True, False)
                else:
                    shift_flags |= np.logical_and(
                                   self.eta[i0:i1] >= region[0],
                                   self.eta[i0:i1] < region[1])

            shift1_zero = self.shift1[n][segment]
            if extrapolate:
                self.xi_corr[i0:i1] = np.where(shift_flags,
                        self.xi_corr[i0:i1] - shift1_zero,
                        self.xi_corr[i0:i1])
            else:
                # Note that i0 & i1 do not necessarily correspond to t0 and t1.
                t0 = self.lamp_median[n]
                t1 = self.lamp_median[n+1]
                if t1 <= t0:
                    slope = 0.
                else:
                    slope =(self.shift1[n+1][segment] -
                            self.shift1[n][segment]) / (t1 - t0)
                self.xi_corr[i0:i1] = np.where(shift_flags,
                    self.xi_corr[i0:i1] -
                        ((self.time[i0:i1] - t0) * slope + shift1_zero),
                    self.xi_corr[i0:i1])

    def findFlash(self, output=None, update=True):
        """Find tagflash wavecals in science data.

        Parameters
        ----------
        output: str or None
            If specified, write array of count rates to this file
            (for testing).

        update: boolean
            If True, keywords in input file will be updated.
        """

        detector = self.info["detector"]
        life_adj_offset = self.info["life_adj_offset"]  # usually 0.


        xtractab = self.reffiles["xtractab"]

        filter = {"opt_elem": self.info["opt_elem"],
                  "cenwave": self.info["cenwave"],
                  "aperture": "WCA"}

        # src_low to src_high is a range for getting the source count rates,
        # which is done by ccos.getbkgcounts.
        if detector == "FUV":
            filter["segment"] = self.info["segment"]
            xtract_info = cosutil.getTable(xtractab, filter, exactly_one=True)
            b_spec = xtract_info.field("b_spec")[0] + life_adj_offset
            height = xtract_info.field("height")[0]
            src_low = int(round(b_spec - height))
            src_high = int(round(b_spec + height))
            eta = self.events.field("ycorr")
        elif self.info["obstype"] == "IMAGING":
            b_spec = Y0 + life_adj_offset
            height = 2 * DY
            src_low = int(round(b_spec - height))
            src_high = NUV_Y - 1
            eta = self.events.field("rawy")
        else:
            # For NUV use only the data for stripe A.
            filter["segment"] = "NUVA"
            xtract_info = cosutil.getTable(xtractab, filter, exactly_one=True)
            b_spec = xtract_info.field("b_spec")[0] + life_adj_offset
            height = xtract_info.field("height")[0]
            src_low = int(round(b_spec - height))
            # Include source counts for wavecal stripes NUVB and NUVC as well.
            src_high = NUV_Y - 1
            eta = self.events.field("rawy")

        # Dummy values, so no background counts will be found.
        bkg1_low = -10
        bkg1_high = -20
        bkg2_low = -10
        bkg2_high = -20
        bkgsf = 0.

        # Get time as a single precision array, for ccos.getstartstop.
        time = cosutil.getColCopy(data=self.events, column="time")
        dq = np.zeros(len(time), dtype=np.int16)
        nbins = int(math.ceil((np.float64(time[-1]) - np.float64(time[0])) / self.delta_t))
        istart = np.zeros(nbins, dtype=np.int32)
        istop = np.zeros(nbins, dtype=np.int32)
        src_counts = np.zeros(nbins, dtype=np.int32)
        bkg_counts = np.zeros(nbins, dtype=np.int32)

        ccos.getstartstop(time, istart, istop, self.delta_t)
        ccos.getbkgcounts(eta, dq, istart, istop,
                          bkg_counts, src_counts,
                          bkg1_low, bkg1_high, bkg2_low, bkg2_high,
                          src_low, src_high, bkgsf)
        del dq, bkg_counts

        # Convert to count rate.
        self.src_counts = src_counts.astype(np.float64) / self.delta_t
        del src_counts

        if output is not None:
            ofd = open(output, "w")
            for val in self.src_counts:
                ofd.write("%g\n" % val)
            ofd.close()

        (hist, step) = self.makeHistogram()

        if hist is None:
            self.numflash = 0
        else:
            cutoff = self.findCutoff(hist, step)
            self.findLampOn(cutoff, time[0])
            self.findLampMedian()
        self.lamp_is_on = (self.numflash > 0)

        if cosutil.checkVerbosity(VERBOSE):
            self.printInfo()

        if update:
            self.updateHeader()

    def makeHistogram(self):
        """Make a histogram of src_counts.

        The function value is (hist, step), the histogram and step size.
        If the maximum value in src_counts is less than 20, the count
        rate is so low that it would be difficult to determine when the
        lamp turned on or off, and (None, None) will be returned to
        indicate this case.

        Returns
        -------
        tuple of array_like and float
            The histogram (array) and step size (float).
        """

        maxval = np.maximum.reduce(self.src_counts)
        if maxval <= MIN_COUNT_RATE:
            return (None, None)

        i_maxval = int(round(maxval))
        nbins = max(20, len(self.src_counts) // 100)
        step = maxval / float(nbins)
        hist = np.zeros(nbins, dtype=np.int32)
        for src in self.src_counts:
            i = int(src / step)
            if i < nbins:               # ignore max value
                hist[i] += 1
        cosutil.printMsg("tagflash histogram = %s" % repr(hist),
                         VERY_VERBOSE)
        cosutil.printMsg("step size for histogram = %g" % step, VERY_VERBOSE)

        return (hist, step)

    def findCutoff(self, hist, step):
        """Find the count rate above which the lamp is probably on.

        The histogram is likely to have two peaks, one close to zero that
        is due to dark counts and one close to the maximum count rate
        that is due to the wavecal lamp being on.  This function searches
        for the second one by looking for the maximum of the upper half
        (higher count rates) of the histogram.  Then the cutoff is set
        to half the count rate corresponding to that peak.

        Parameters
        ----------
        hist: array_like
            Histogram of count rates src_counts.  hist[i] is the
            number of elements of src_counts with count rates between
            i*step and (i+1)*step.

        step: float
            Step size for histogram, i.e. change in count rate
            from hist[i] to hist[i+1].

        Returns
        -------
        float
            The cutoff count rate.
        """

        n = len(hist)
        istart = n // 2
        index = istart
        maxhist = 0
        for i in range(n // 2, n):
            if hist[i] > maxhist:
                maxhist = hist[i]
                index = i

        cutoff = (index / 2.) * step
        cosutil.printMsg("tagflash cutoff = %.2f counts/s" % cutoff,
                         VERY_VERBOSE)

        return cutoff

    def findLampOn(self, cutoff, t0):
        """Find the actual times when the lamps turned on and off.

        The nominal lamp turn-on and turn-off times were gotten from
        the EVENTS header and saved as self.lamp_on and self.lamp_off.
        This function gets the actual times, constrained to be within
        the time intervals specified in the header.  The attributes
        lamp_on and lamp_off are then updated.  The number of flashes,
        the numflash attribute, can also be modified.

        Parameters
        ----------
        cutoff: float
            A count rate (in src_counts array) greater than this
            indicates that the wavecal lamp was on.

        t0: float
            Time of first photon event (first element of TIME column).
        """

        nbins = len(self.src_counts)

        # Find the turn-on and turn-off time of each flash, starting with
        # the nominal times in self.lamp_on and self.lamp_off.
        lamp_on = []
        lamp_off = []

        for i in range(self.numflash):
            lamp_is_on = False
            lamp_on_i = self.lamp_on[i] - self.buffer_on
            lamp_off_i = self.lamp_off[i] + self.buffer_off
            (k_on, k_off) = self.getIndices(nbins, t0, lamp_on_i, lamp_off_i)
            if k_on is None:
                continue
            # search for the actual lamp turn-on time
            for k in range(k_on, k_off+1):
                if self.src_counts[k] > cutoff:
                    lamp_is_on = True
                    t = np.float64(t0) + k * self.delta_t
                    lamp_on.append(t)
                    break
            if not lamp_is_on:
                continue
            # search for the actual lamp turn-off time
            for k in range(k_off, k_on-1, -1):
                if self.src_counts[k] > cutoff:
                    t = np.float64(t0) + (k+1) * self.delta_t
                    t = min(t, self.time[-1])
                    lamp_off.append(t)
                    break
        if len(lamp_on) != len(lamp_off):
            raise RuntimeError("Internal error:  len(lamp_on) = %d, "
                               "len(lamp_off) = %d" % \
                               (len(lamp_on), len(lamp_off)))

        self.numflash = len(lamp_on)            # the actual number of flashes

        self.lamp_on = lamp_on
        self.lamp_off = lamp_off

    def findLampMedian(self):
        """Find the median time of each wavecal flash."""

        lamp_median = []
        for i in range(self.numflash):
            (i0, i1) = ccos.range(self.time, self.lamp_on[i], self.lamp_off[i])
            index = (i0 + i1) // 2
            lamp_median.append(self.time[index])

        self.lamp_median = lamp_median

    def getIndices(self, nbins, t0, lamp_on_i, lamp_off_i):
        """Compute indices in src_counts corresponding to given times.

        Parameters
        ----------
        nbins: int
            Number of elements in src_counts array.

        t0: float
            Time of first photon event.

        lamp_on_i: float
            One element of lamp_on array (minus buffer_on).

        lamp_off_i: float
            One element of lamp_off array (plus buffer_off).

        Returns
        -------
        tuple of two ints
            The indices corresponding to lamp_on_i and lamp_off_i;
            both values will be None if these times are outside the
            interval (t0, t0+delta_t).
        """

        k_on = (lamp_on_i - np.float64(t0)) / self.delta_t
        k_off = (lamp_off_i - np.float64(t0)) / self.delta_t
        k_on = int(math.floor(k_on))
        k_off = int(math.ceil(k_off))
        if k_on >= nbins or k_off < 0:
            k_on = None
            k_off = None
        else:
            k_on = max(k_on, 0)
            k_off = min(k_off, nbins-1)

        return (k_on, k_off)

    def printInfo(self):
        """Print time info for each wavecal flash."""

        if self.numflash < 1:
            return

        cosutil.printMsg("lamp on, off, duration, median time:")
        for i in range(self.numflash):
            cosutil.printMsg("%d:  %.1f  %.1f  %.1f  %.1f" %
                             (i+1, self.lamp_on[i], self.lamp_off[i],
                             self.lamp_off[i] - self.lamp_on[i],
                             self.lamp_median[i]))

    def updateHeader(self):
        """Assign extension header keywords with updated info.

        This function updates keywords with actual values (the input
        header contained commanded values).  The keywords that will be
        updated are NUMFLASH, LMP_ONi, LMPOFFi, LMPDURi, and LMPMEDi;
        i normally runs from 1 to NUMFLASH inclusive, but if NUMFLASH is
        large not all the keywords will be present in the input header,
        and in this case (the test is on LMP_ONi) the remaining values
        will not be written to the header.  Note that the _tagflash.fits
        table does contain these values for every flash.
        """

        self.hdr["NUMFLASH"] = self.numflash

        for i in range(self.numflash):
            keyword = "LMP_ON%d" % (i+1)
            if keyword not in self.hdr:
                break
            self.hdr[keyword] = self.lamp_on[i]
            keyword = "LMPOFF%d" % (i+1)
            self.hdr[keyword] = self.lamp_off[i]
            keyword = "LMPDUR%d" % (i+1)
            self.hdr[keyword] = self.lamp_off[i] - self.lamp_on[i]
            keyword = "LMPMED%d" % (i+1)
            self.hdr[keyword] = self.lamp_median[i]

    def avgShift(self):
        """Compute the shift1 and shift2 offsets averaged over time.

        Returns
        -------
        tuple of two dictionaries
            The key is the segment or stripe name.  The dictionaries
            contain the shifts in the dispersion and cross-dispersion
            directions averaged over the exposure.
        """

        avg_dx = {}
        avg_dy = {}
        if self.numflash < 1 or self.info["exptime"] <= 0.:
            for segment in self.segment_list:
                avg_dx[segment] = 0.
                avg_dy[segment] = 0.
            return (avg_dx, avg_dy)

        time = self.time
        # Note that this is the entire time interval, i.e. including any
        # bad time intervals.
        time_range = time[-1] - time[0]
        if time_range <= 0.:
            time_range = 1.

        for segment in self.segment_list:
            t_prev = time[0]
            shift1_prev = self.shift1[0][segment]
            shift2_prev = self.shift2[0]
            sum_t = 0.
            sum_shift1 = 0.
            sum_shift2 = 0.
            for n in range(self.numflash):
                t = self.lamp_median[n]
                shift1 = self.shift1[n][segment]
                shift2 = self.shift2[n]
                sum_shift1 += (t - t_prev) * (shift1 + shift1_prev) / 2.
                sum_shift2 += (t - t_prev) * (shift2 + shift2_prev) / 2.
                t_prev = t
                shift1_prev = shift1
                shift2_prev = shift2

            if time[-1] > t:
                sum_shift1 += (time[-1] - t) * shift1
                sum_shift2 += (time[-1] - t) * shift2
            avg_dx[segment] = sum_shift1 / time_range
            avg_dy[segment] = sum_shift2 / time_range

        return (avg_dx, avg_dy)

    def setShiftKeywords(self, avg_dx, avg_dy):
        """Assign values to the shift keywords.

        Keywords SHIFT1[ABC] will be set to the average offset in the
        dispersion direction, SHIFT2[ABC] to the average offset in the
        cross-dispersion direction, and DPIXEL1[ABC] to:
            XFULL - (XFULL rounded to an integer)
        Keyword values will be set in the corrtag header and in the output
        lampflash header.

        LAMPUSED may also be updated.  This keyword can be NONE in the raw
        header if the info for the support file was taken at times when the
        lamp happened to be off.  If the lamp really was on, the keyword
        will be set to the value of LAMPPLAN.

        Parameters
        ----------
        avg_dx: dictionary
            The average shift in the dispersion direction; the key is the
            segment or stripe name.

        avg_dy: dictionary
            The average shift in the cross-dispersion direction; the key is
            the segment or stripe name.
        """

        for segment in self.segment_list:
            key = "SHIFT1" + segment[-1]
            value = round(avg_dx[segment], 4)
            self.hdr[key] = value                               # corrtag
            self.ofd[1].header[key] = value                     # lampflash
            key = "SHIFT2" + segment[-1]
            value = round(avg_dy[segment], 4)
            self.hdr[key] = value
            self.ofd[1].header[key] = value
            sum_chisq = 0.
            sum_ndf = 0
            for n in range(self.numflash):
                sum_chisq += self.chi_square[n][segment]
                sum_ndf += (self.n_deg_freedom[n][segment] + 1)
            sum_ndf -= 1
            if self.override_segment_B:
                sum_ndf = 0
            key = "chi_sq_" + segment[-1]
            self.hdr[key] = round(sum_chisq, 1)
            self.ofd[1].header[key] = round(sum_chisq, 1)
            key = "ndf_" + segment[-1]
            self.hdr[key] = sum_ndf
            self.ofd[1].header[key] = sum_ndf

            # use self.regions for dpixel1[abc]
            shift_flags = np.zeros(len(self.eta), dtype=np.bool_)
            locn_list = self.regions[segment]
            # if NUV, take the region for the PSA (lower pixel numbers)
            region = locn_list[0]
            if region[0] is None:
                shift_flags |= np.where(self.eta < region[1], True, False)
            elif region[1] is None:
                shift_flags |= np.where(self.eta >= region[0], True, False)
            else:
                shift_flags |= np.logical_and(self.eta >= region[0],
                                              self.eta < region[1])
            xi = self.xi_corr[shift_flags]      # copy out the relevant subset
            if len(xi) > 0:
                xi_diff = xi - np.around(xi)
                dpixel1 = xi_diff.mean(dtype=np.float64)
                value = round(dpixel1, 4)
            else:
                value = 0.
            key = "DPIXEL1" + segment[-1]
            self.hdr[key] = value
            self.ofd[1].header[key] = value

        if self.override_segment_B:
            lampused = self.ofd[0].header.get("lampused", "missing")
            if lampused != "missing":
                self.phdr["lampused"] = lampused
        else:
            lampused = self.phdr.get("lampused", "missing")
            lampplan = self.phdr.get("lampplan", "missing")
            if self.lamp_is_on and lampused == "NONE":
                if lampplan == "missing":
                    cosutil.printWarning("The wavecal lamp was on, " \
                                         "but LAMPUSED = %s and LAMPPLAN is " \
                                         "missing." % lampused, level=VERBOSE)
                else:
                    cosutil.printMsg("LAMPUSED = %s, which is incorrect; " \
                                     "the value will be reset to %s." % \
                                     (lampused, lampplan), level=VERBOSE)
                    self.phdr["lampused"] = lampplan
                    self.ofd[0].header["lampused"] = lampplan
            if not self.lamp_is_on:
                cosutil.printWarning("The wavecal lamp was not on " \
                                     "for tagflash data.", level=VERBOSE)
                if lampused != "NONE":
                    cosutil.printMsg("LAMPUSED = %s, and it will be reset " \
                                     "to NONE." % lampused, level=VERBOSE)
                    self.phdr["lampused"] = "NONE"
                    self.ofd[0].header["lampused"] = "NONE"

    def shift1VsTime(self):
        """Interpolate shift1 at one-second intervals.

        Returns
        -------
        (tl_time, shift1_vs_time): tuple of two array like
            tl_time is the array of times at one-second intervals, for
            the timeline table.  shift1_vs_time is the array of
            corresponding values of shift1a or shift1b, or None if there
            were no flashes or if the exposure time is zero.
        """

        if self.numflash < 1 or self.info["exptime"] <= 0.:
            return (cosutil.timelineTimes(None, 0.), None)

        first_time = self.time[0]
        tl_time = cosutil.timelineTimes(self.time[0], self.time[-1], dt=1.)
        nbins = len(tl_time)
        shift1_vs_time = np.zeros(nbins, dtype=np.float32)

        if self.segment_list[0][0] == "F":
            segment = self.segment_list[0]
        elif "NUVB" in self.segment_list:
            segment = "NUVB"
        else:
            segment = self.segment_list[0]

        for n in range(self.numflash):

            t0 = self.lamp_median[n]
            if n == self.numflash - 1:
                t1 = tl_time[-1]
            else:
                t1 = self.lamp_median[n+1]

            shift1_zero = self.shift1[n][segment]
            t0 = self.lamp_median[n]
            i0 = int(round(t0 - first_time))
            i0 = max(i0, 0)

            if self.numflash == 1:
                shift1_vs_time[:] = shift1_zero
            elif n == self.numflash - 1:
                # last flash; extrapolate with zero slope
                shift1_vs_time[i0:] = shift1_zero
            else:
                if n == 0:      # extrapolate with slope of first interval
                    i0 = 0
                t1 = self.lamp_median[n+1]
                if t1 <= t0:
                    slope = 0.
                else:
                    slope = (self.shift1[n+1][segment] -
                             self.shift1[n][segment]) / (t1 - t0)
                i1 = int(round(t1 - first_time))
                i1 = min(i1, nbins-1)
                shift1_vs_time[i0:i1] = \
                        ((tl_time[i0:i1] - t0) * slope + shift1_zero)

        return (tl_time, shift1_vs_time)

class FUVConcurrentWavecal(ConcurrentWavecal):

    def __init__(self, events, outflash, shift_file,
                 info, switches, reffiles, phdr, hdr):

        ConcurrentWavecal.__init__(self, events, outflash, shift_file,
                                   info, switches, reffiles, phdr, hdr)
        self.xi  = events.field("XDOPP")
        self.eta = events.field("YCORR")
        self.dq  = events.field("DQ")
        self.xi_corr  = events.field("XFULL")
        self.eta_corr = events.field("YFULL")
        self.spectrum = np.zeros(FUV_EXTENDED_X, dtype=np.float64)
        self.segment_list = [info["segment"]]
        self.override_segment_B = cosutil.checkForNoWavecalData(
                        info["opt_elem"], info["cenwave"], info["segment"],
                        reffiles["lamptab"])
        if self.override_segment_B:
            if outflash.endswith("_b.fits"):
                index = outflash.rfind("b.fits")
                self.lampflash_a = outflash[0:index] + "a.fits"
                if os.access(self.lampflash_a, os.R_OK):
                    cosutil.printMsg("Info:  No wavecal signal for FUVB, "
                                     " so info will be copied from FUVA.")
                    self.segment_A_present = True
                else:
                    cosutil.printMsg("Info:  No wavecal signal for FUVB,"
                                     " but the file")
                    cosutil.printContinuation(
                        "%s for segment A does not exist,"
                        % self.lampflash_a)
                    cosutil.printContinuation(
                        "so wavecal shifts will be set to the default.")
                    self.segment_A_present = False
            else:
                cosutil.printWarning("No wavecal signal, "
                                     "but don't understand the name")
                cosutil.printContinuation("outflash = %s," % outflash)
                cosutil.printContinuation(
                        "so wavecal shifts will be set to the default.")
                self.segment_A_present = False

        # Copy xi and eta to the columns for corrected values.
        self.copyColumns()

        # The shift1 offset should be applied only within this region.
        (b_low, b_high, b_left, b_right) = \
                cosutil.activeArea(info["segment"], reffiles["brftab"])
        self.regions[info["segment"]] = [(b_low, b_high)]

    def copySegAtoSegB(self):
        """Copy lampflash_a.fits to lampflash_b.fits.

        This method is called for the case of FUVB where there is no
        wavecal data but there is data for segment A.  The lampflash file
        for segment A will be copied to <rootname>_lampflash_b.fits.
        The latter file will be opened (with FITS), and the "FUVA" in
        the SEGMENT column will be changed to "FUVB"; the file should
        be closed later by calling closeOutFlash().
        """

        # copy rootname_lampflash_a.fits to rootname_lampflash_b.fits
        cosutil.copyFile(self.lampflash_a, self.outflash)

        self.ofd = fits.open(self.outflash, mode="update")
        self.ofd[0].header["segment"] = self.segment_list[0]
        if self.ofd[1].data is None or len(self.ofd[1].data) == 0:
            self.numflash = 0
            cosutil.printWarning("No data in lampflash table.")
            self.phdr["wavecorr"] = "SKIPPED"
            return

        # A_shift1 is the shift1 value for FUVA, and A_spec_found is used
        # for printing "not found in FUVA" if the shift was not found.
        A_shift1 = self.ofd[1].data.field("shift_disp").copy()
        A_shift1[:] = A_shift1 + DELTA_SHIFT1
        A_spec_found = self.ofd[1].data.field("spec_found").copy()

        nrows = len(self.ofd[1].data)
        # This assumes that lampflash_a contains data for only one segment.
        self.numflash = nrows
        self.lamp_median = self.ofd[1].data.field("time")
        self.shift2 = self.ofd[1].data.field("shift_xdisp")
        self.spec_found = self.ofd[1].data.field("spec_found")

        segment_column = self.ofd[1].data.field("segment")
        shift1 = self.ofd[1].data.field("shift_disp")
        chi_square = self.ofd[1].data.field("chi_square")
        n_deg_freedom = self.ofd[1].data.field("n_deg_freedom")

        # Create lists of the appropriate length.
        self.shift1 = [0] * nrows
        self.chi_square = [0] * nrows
        self.n_deg_freedom = [0] * nrows
        # The elements of the lists are dictionaries.
        segment = self.segment_list[0]
        for i in range(len(shift1)):
            self.shift1[i] = {segment: A_shift1[i]}
            self.chi_square[i] = {segment: 0.}
            self.n_deg_freedom[i] = {segment: 0}

        segment_column[:] = self.segment_list[0]    # replace FUVA with FUVB
        self.shift2[:] += DELTA_SHIFT2
        self.spec_found[:] = False

        # Override shifts if specified in shift_file, and print shift info.
        row = 0
        cosutil.printMsg(
"  segment    cross-disp        dispersion direction", VERBOSE)
        cosutil.printMsg(
"            shift (locn)      shift  [orig.]  chi sq (n)", VERBOSE)
        cosutil.printMsg(
"  -------   -------------     --------------------------", VERBOSE)
        for n in range(self.numflash):
            for segment in self.segment_list:
                if segment != segment_column[row]:
                    cosutil.printWarning("Out of synch at row %d in %s!" % \
                                         (row, self.outflash))
                    return
                user_specified = False
                if self.user_shifts is not None:        # override shifts?
                    # note that flash number is one indexed
                    ((user_shift1, user_shift2), nfound) = \
                        self.user_shifts.getShifts((n+1, segment))
                    if user_shift1 is None:
                        user_specified = False
                    else:
                        user_specified = True
                        self.shift1[row] = {segment: user_shift1}
                message = "%2d %4s %9.1f (-999) %9.1f [%5.1f]  %6.1f (%d)" \
                            % (n+1, segment, self.shift2[row],
                               self.shift1[row][segment], A_shift1[row], 0., 0)
                if user_specified:
                    message = message + "  # user-specified"
                elif not A_spec_found[row]:
                    message = message + "  # not found in FUVA"
                else:
                    message = message + "  # based on FUVA value"
                cosutil.printMsg(message, VERBOSE)
                row += 1

        self.phdr["wavecorr"] = "COMPLETE"

    def setSegBtoZero(self):
        """Set the shifts to zero (no segment A data to copy)."""

        cosutil.printMsg(
"  segment    cross-disp           dispersion direction", VERBOSE)
        cosutil.printMsg(
"            shift (locn)      shift err  [orig.]    FP   chi sq (n)", VERBOSE)
        cosutil.printMsg(
"  -------   -------------     -------------------------  ----------", VERBOSE)
        n = 0
        segment = self.segment_list[0]
        user_specified = False          # may be reset below
        self.numflash = 1               # there might not have been any
        # Create one-element lists.
        self.lamp_on = [0.]
        self.lamp_off = [0.]
        self.lamp_duration = [0.]
        self.lamp_median = [0.]
        self.shift2 = [0.]
        self.spec_found = [False]
        self.chi_square = [{segment: 0.}]
        self.n_deg_freedom = [{segment: 0}]

        # get fp_pixel_shift for shift1
        lamptab = self.reffiles["lamptab"]
        filter_lamp = {"opt_elem": self.info["opt_elem"],
                       "cenwave": self.info["cenwave"],
                       "segment": segment}
        got_pixel_shift = cosutil.findColumn(lamptab, "fp_pixel_shift")
        if got_pixel_shift:
            filter_lamp["fpoffset"] = self.info["fpoffset"]
        lamp_info = cosutil.getTable(lamptab, filter_lamp)
        if lamp_info is not None and got_pixel_shift:
            fp_pixel_shift = lamp_info.field("fp_pixel_shift")[0]
        else:
            fp_pixel_shift = 0.
        # this may be replaced below, if the user specified the value
        self.shift1 = [{segment: fp_pixel_shift}]

        if self.user_shifts is not None:
            # Check for a user-supplied value for shift1.
            # Note that flash number is one indexed.
            ((user_shift1, user_shift2), nfound) = \
                self.user_shifts.getShifts((n+1, segment))
            if user_shift1 is None:
                user_specified = False
            else:
                user_specified = True
                self.shift1 = [{segment: user_shift1}]
        message = \
"%2d %4s %9.1f (-9999) %9.1f 0.00 [  0.0] %6.1f     0.0 (0)" \
                    % (n+1, segment, self.shift2[n],
                       self.shift1[n][segment], fp_pixel_shift)
        if user_specified:
            message = message + "  # user-specified"
            # Not skipped because the user specified the shift.
            self.wavecorr = "COMPLETE"
        else:
            message = message + "  # set to 0"
            self.wavecorr = "SKIPPED"
        cosutil.printMsg(message, VERBOSE)

    def miscSegB(self):
        """Miscellaneous stuff for FUVB, with no segment A data."""

        segment = self.segment_list[0]

        # compute wavelengths
        disptab = self.reffiles["disptab"]
        filter_disp = {"opt_elem": self.info["opt_elem"],
                       "cenwave": self.info["cenwave"],
                       "fpoffset": self.info["fpoffset"],
                       "segment": segment,
                       "aperture": "WCA"}
        pixel = np.arange(len(self.spectrum), dtype=np.float64)
        disp_rel = dispersion.Dispersion(disptab, filter_disp)
        wavelength = disp_rel.evalDisp(pixel)
        disp_rel.close()

        # note that the FITS HDUList object self.ofd must exist at this time
        segment_col = self.ofd[1].data.field("segment")
        wavelength_col = self.ofd[1].data.field("wavelength")
        shift_disp_col = self.ofd[1].data.field("shift_disp")
        spec_found_col = self.ofd[1].data.field("spec_found")
        n = 0
        segment_col[n] = segment
        wavelength_col[n][:] = wavelength
        shift_disp_col[n] = self.shift1[n][segment]
        spec_found_col[n] = False

        avg_dx = {segment: self.shift1[0][segment]}
        avg_dy = {segment: 0.}
        return (avg_dx, avg_dy)

    def closeOutFlash(self):
        """Close the lampflash_b.fits file."""

        self.ofd.close()
        self.ofd = None

    def shift2Corr(self, n, i0, i1, extrapolate=False):
        """Correct the pixel coordinates in the cross-dispersion direction.

        The difference between this version and the one for NUV is that this
        one limits the shift to the active area.

        Parameters
        ----------
        n: int
            Apply shift2 for the nth time interval between wavecals.

        i0: int
            [i0:i1] is the slice of event numbers to be corrected.

        i1: int
            [i0:i1] is the slice of event numbers to be corrected.

        extrapolate: boolean
            True if n is the last flash.
        """

        # Restrict the correction to the applicable region.  Note that the
        # limits of the region (the active area) are not adjusted by shift2.
        shift_flags = np.zeros(i1 - i0, dtype=np.bool_)
        region = self.regions[self.segment_list[0]][0]
        shift_flags |= np.logical_and(self.eta[i0:i1] >= region[0],
                                      self.eta[i0:i1] <= region[1])

        shift2_zero = self.shift2[n]
        if extrapolate:
            self.eta_corr[i0:i1] = np.where(shift_flags,
                                            self.eta_corr[i0:i1] - shift2_zero,
                                            self.eta_corr[i0:i1])
        else:
            # Note that i0 & i1 do not necessarily correspond to t0 and t1,
            # because we can extrapolate to the beginning or end of the array.
            t0 = self.lamp_median[n]
            t1 = self.lamp_median[n+1]
            if t1 <= t0:
                slope = 0.
            else:
                slope = (self.shift2[n+1] - self.shift2[n]) / (t1 - t0)
            self.eta_corr[i0:i1] = np.where(shift_flags,
                        self.eta_corr[i0:i1] -
                            ((self.time[i0:i1] - t0) * slope + shift2_zero),
                        self.eta_corr[i0:i1])

class NUVConcurrentWavecal(ConcurrentWavecal):

    def __init__(self, events, outflash, shift_file,
                 info, switches, reffiles, phdr, hdr):

        ConcurrentWavecal.__init__(self, events, outflash, shift_file,
                                   info, switches, reffiles, phdr, hdr)
        self.xi  = events.field("XDOPP")
        self.eta = events.field("RAWY")
        self.dq  = events.field("DQ")
        self.xi_corr  = events.field("XFULL")
        self.eta_corr = events.field("YFULL")
        self.spectrum = np.zeros(NUV_EXTENDED_X, dtype=np.float64)
        self.segment_list = ["NUVA", "NUVB", "NUVC"]

        # Copy xi and eta to the columns for corrected values.
        self.copyColumns()

        # The shift1 offset should be applied only within this region.
        self.regions = self.setRegions()

    def setRegions(self):
        """Determine the regions over which shift1 should be applied.

        The function value is a dictionary with nominally three entries.
        Segment name is the key, and each value is a list of the two
        intervals (one for PSA, one for WCA) over which the shift in the
        dispersion direction (shift1) should be applied.  The limits of
        the intervals are the midpoints between b_spec values from the
        xtractab.  Note that shift2 has not been subtracted from these
        intervals, and it should not be, because the cross-dispersion
        positions will be corrected for shift2 before correcting the
        positions in the dispersion direction.
        """

        # segment and aperture will be added to the filter in the loop.
        filter = {"opt_elem": self.info["opt_elem"],
                  "cenwave": self.info["cenwave"]}

        # locations will be a list of tuples, each containing the
        # segment name and nominal location.
        middle = float(NUV_X) / 2.
        locations = []
        for segment in self.segment_list:

            filter["segment"] = segment

            filter["aperture"] = "WCA"
            xtract_info = cosutil.getTable(self.reffiles["xtractab"], filter,
                                           at_least_one=True)
            # Note that this does not include life_adj_offset; see the comment
            # in the doc string about shift2.
            b_spec = xtract_info.field("b_spec")[0] + \
                     xtract_info.field("slope")[0] * middle
            locations.append((segment, b_spec))

            filter["aperture"] = "PSA"
            xtract_info = cosutil.getTable(self.reffiles["xtractab"], filter,
                                           at_least_one=True)
            b_spec = xtract_info.field("b_spec")[0] + \
                     xtract_info.field("slope")[0] * middle
            locations.append((segment, b_spec))

        locations.sort(key=r_key)       # sort on b_spec, regardless of segment
        len_locn = len(locations)

        # intervals will be the same length as locations.  Each interval
        # [first,last] is the slice over which shift1[a-c] should be
        # applied.  There should be six such intervals, one for each
        # stripe, and for apertures PSA and WCA.
        # Use None for the lower and upper cutoffs so every event will be
        # included.
        intervals = []
        first = None                    # no lower cutoff
        for i in range(len_locn):
            segment = locations[i][0]
            locn = locations[i][1]
            if i == len_locn - 1:
                last = None             # no upper cutoff
            else:
                next_locn = locations[i+1][1]
                # midpoint between adjacent b_spec values
                last = int(round((locn + next_locn) / 2.))
            intervals.append([segment, [first,last]])
            first = last

        regions = {}
        for segment in self.segment_list:
            locn_list = []
            for i in range(len_locn):
                if segment == intervals[i][0]:
                    locn_list.append(intervals[i][1])
            regions[segment] = locn_list

        return regions

    def shift2Corr(self, n, i0, i1, extrapolate=False):
        """Correct the pixel coordinates in the cross-dispersion direction.

        Parameters
        ----------
        n: int
            Apply shift2 for the nth time interval between wavecals.

        i0: int
            [i0:i1] is the slice of event numbers to be corrected.

        i1: int
            [i0:i1] is the slice of event numbers to be corrected.

        extrapolate: boolean
            True if n is the last flash.
        """

        shift2_zero = self.shift2[n]
        if extrapolate:
            self.eta_corr[i0:i1] -= shift2_zero
        else:
            # Note that i0 & i1 do not necessarily correspond to t0 and t1,
            # because we can extrapolate to the beginning or end of the array.
            t0 = self.lamp_median[n]
            t1 = self.lamp_median[n+1]
            if t1 <= t0:
                slope = 0.
            else:
                slope = (self.shift2[n+1] - self.shift2[n]) / (t1 - t0)
            self.eta_corr[i0:i1] -= \
                        ((self.time[i0:i1] - t0) * slope + shift2_zero)

class NUVImagingWavecal(ConcurrentWavecal):

    def __init__(self, events, outflash, shift_file,
                 info, switches, reffiles, phdr, hdr):

        info_copy = copy.deepcopy(info)
        info_copy["cenwave"] = 0

        ConcurrentWavecal.__init__(self, events, outflash, shift_file,
                                   info_copy, switches, reffiles, phdr, hdr)
        self.xi  = events.field("XDOPP")
        self.eta = events.field("RAWY")
        self.dq  = events.field("DQ")
        self.xi_corr  = events.field("XFULL")
        self.eta_corr = events.field("YFULL")
        self.spectrum = np.zeros(NUV_X, dtype=np.float64)
        self.segment_list = ["N/A"]

        # Copy xi and eta to the columns for corrected values.
        self.copyColumns()

        # The shift1 offset should be applied over the entire detector.
        self.regions = self.setRegions()

    def setRegions(self):
        """Shift1 should be applied over the full detector.

        The function value is a dictionary with one entry.  The key is the
        "stripe" name "NUVA", and the value (there's only one) is a list of
        lists of the interval (the full detector) over which the shift in the
        dispersion direction (shift1) should be applied.
        """

        regions = {}
        for segment in self.segment_list:
            regions[segment] = [[0, NUV_Y]]

        return regions

    def shift2Corr(self, n, i0, i1, extrapolate=False):
        """Correct the pixel coordinates in the cross-dispersion direction.

        n: int
            Apply shift2 for the nth time interval between wavecals.

        i0: int
            [i0:i1] is the slice of event numbers to be corrected.

        i1: int
            [i0:i1] is the slice of event numbers to be corrected.

        extrapolate: boolean
            True if n is the last flash.
        """

        shift2_zero = self.shift2[n]
        if extrapolate:
            self.eta_corr[i0:i1] -= shift2_zero
        else:
            # Note that i0 & i1 do not necessarily correspond to t0 and t1,
            # because we can extrapolate to the beginning or end of the array.
            t0 = self.lamp_median[n]
            t1 = self.lamp_median[n+1]
            if t1 <= t0:
                slope = 0.
            else:
                slope = (self.shift2[n+1] - self.shift2[n]) / (t1 - t0)
            self.eta_corr[i0:i1] -= \
                        ((self.time[i0:i1] - t0) * slope + shift2_zero)

    def findShifts(self):
        """For each wavecal flash, find the shift in each axis."""

        fpoffset = self.info["fpoffset"]
        segment = self.segment_list[0]          # don't really need segment

        # Find the offsets in both axes, for each wavecal exposure.
        cosutil.printMsg(
"         Y direction       X direction", VERBOSE)
        cosutil.printMsg(
"        shift (locn)      shift (locn)", VERBOSE)
        cosutil.printMsg(
"        ------------      ------------", VERBOSE)
        row = 0
        for n in range(self.numflash):
            (i0, i1) = ccos.range(self.time, self.lamp_on[n], self.lamp_off[n])

            shift1 = {}                 # to be saved in an attribute
            chi_square = {}             # to be saved in an attribute
            n_deg_freedom = {}          # to be saved in an attribute
            # First find the median of the xi and eta positions.
            nelem = i1 - i0
            index_x = self.xi[i0:i1].argsort()
            index_y = self.eta[i0:i1].argsort()
            x_median = self.xi[i0+index_x[nelem//2]]
            y_median = self.eta[i0+index_y[nelem//2]]

            # Now find the mean values of xi and of eta, but restrict
            # the range to the median plus or minus DX or DY.
            select = np.zeros(len(self.xi), dtype=np.bool_)
            select[i0:i1] = 1
            select = np.where(self.xi < x_median-DX, False, select)
            select = np.where(self.xi > x_median+DX, False, select)
            select = np.where(self.eta < y_median-DY, False, select)
            select = np.where(self.eta > y_median+DY, False, select)
            select = select.astype(np.bool_)

            x = self.xi[select].mean(dtype=np.float64)
            y = self.eta[select].mean(dtype=np.float64)
            shift1[segment] = x - X0
            shift2 = y - Y0
            chi_square[segment] = 0.            # not used
            n_deg_freedom[segment] = 0

            message = "%2d %9.1f (%5.1f) %9.1f (%5.1f)" \
                            % (n+1, shift2, y, shift1[segment], x)
            user_specified = False
            if self.user_shifts is not None:            # override shifts?
                # note that flash number is one indexed
                ((user_shift1, user_shift2), nfound) = \
                    self.user_shifts.getShifts((n+1, segment))
                if user_shift1 is not None:
                    shift1[segment] = user_shift1
                    user_specified = True
                if user_shift2 is not None:
                    shift2 = user_shift2
                    user_specified = True
            if user_specified:
                message = message + "  # user-specified"
            cosutil.printMsg(message, VERBOSE)
            # copy to outflash table data
            self.saveSpectrum(n, row, shift1[segment], shift2, True)
            self.shift1.append(shift1)
            self.shift2.append(shift2)
            self.chi_square.append(chi_square)
            self.n_deg_freedom.append(n_deg_freedom)
            row += 1

    def saveSpectrum(self, n, row, shift1, shift2, spec_found):
        """Copy the spectrum to the record array for the outflash table.

        Parameters
        ----------
        n: int
            Index of current lamp flash.

        row: int
            Row index (zero indexed) in output table.

        shift1: float
            Shift in X direction.

        shift2: float
            Shift in Y direction.

        spec_found: boolean
            Was the wavecal spectrum actually found?
        """

        if self.ofd is None:
            return

        t0 = self.lamp_on[n]
        t1 = self.lamp_off[n]

        self.ofd[1].data.field("time")[row] = self.lamp_median[n]
        self.ofd[1].data.field("exptime")[row] = t1 - t0
        self.ofd[1].data.field("lamp_on")[row] = self.lamp_on[n]
        self.ofd[1].data.field("lamp_off")[row] = self.lamp_off[n]
        self.ofd[1].data.field("shift_disp")[row] = shift1
        self.ofd[1].data.field("shift_xdisp")[row] = shift2
        self.ofd[1].data.field("spec_found")[row] = spec_found

        self.ofd[1].data.field("segment")[row] = "N/A"
        self.ofd[1].data.field("wavelength")[row][:] = 0.
        self.ofd[1].data.field("gross")[row][:] = 0.
        self.ofd[1].data.field("chi_square")[row] = 0.
        self.ofd[1].data.field("n_deg_freedom")[row] = 0



import copy
import math
import shutil
import time
import types
import numpy as np
from numpy import random
from astropy.io import fits

from . import cosutil
from . import ccos
from . import phot
from . import timetag                  # actually for more generic functions
from . import wavecal
from .calcosparam import *       # parameter definitions

def accumBasicCalibration(input, inpha, outtag,
                          outflt, outcounts, outcsum,
                          cl_args,
                          info, switches, reffiles,
                          wavecal_info):
    """Do the basic processing for accum data.

    The function value will be zero if there was no problem.

    Parameters
    ----------
    input: str
        Name of the input file.

    inpha: str
        Name of the input file containing the pulse height histogram
        (FUV only).

    outtag: str
        Name of the output file for pseudo time-tag data.

    outflt: str
        Name of the output file for the flat-fielded count-rate image.

    outcounts: str
        Name of the output file for the count-rate image.

    outcsum: str or None
        Name of the output image for OPUS to add to cumulative image.

    cl_args: dictionary
        Some of the command-line arguments.

    info: dictionary
        Header keywords and values.

    switches: dictionary
        Calibration switches.

    reffiles: dictionary
        Reference file names.

    wavecal_info: list of dictionaries
        When wavecal exposures were processed, the results were stored in
        dictionaries in this list.

    Returns
    -------
    status: int
        0 is OK
        1 means there were no rows in the input table
    """

    cosutil.printIntro("ACCUM calibration")
    if info["exptype"] == "ACQ/IMAGE":
        names = [("Input", input),
                 ("OutFlt", outflt), ("OutCounts", outcounts)]
    else:
        names = [("Input", input), ("pseudo tt", outtag),
                 ("OutFlt", outflt), ("OutCounts", outcounts)]
    if info["detector"] == "FUV":
        names.insert(1, ("InPha", inpha))
    if outcsum is not None:
        names.append(("OutCsum", outcsum))
    cosutil.printFilenames(names,
                           shift_file=cl_args["shift_file"],
                           stimfile=cl_args["stimfile"],
                           livetimefile=cl_args["livetimefile"])
    cosutil.printMode(info)

    if info["corrtag_input"]:
        shutil.copy(input, outtag)
        status = timetag.timetagBasicCalibration(input, inpha, outtag,
                    outflt, outcounts, None, outcsum,
                    cl_args, info, switches, reffiles, wavecal_info)
        return status

    # Get a list of all the headers in the input file.
    headers = cosutil.getHeaders(input)
    phdr = headers[0]
    # Get x_offset now, because overrideKeywords may change it in headers[1].
    x_offset = headers[1].get("x_offset", 0)

    # Update the switches and reference file names, so the output header
    # will reflect what was actually used.
    cosutil.overrideKeywords(phdr, headers[1], info, switches, reffiles)

    # acq/image data are processed differently because they have two imsets.
    if info["exptype"] == "ACQ/IMAGE":
        acqImage(input, outflt, outcounts, outcsum, cl_args,
                 info, switches, reffiles)

    if info["exptype"] == "ACQ/IMAGE":
        return 0

    # Check for null science data.
    if info["npix"] == (0,):
        nrows = 0
    else:
        # Open the accum image.
        # The number of rows in the pseudo time-tag table will be equal to
        # the total number of counts in the input image.
        fd = fits.open(input, mode="copyonwrite")
        sci_hdu = fd[("SCI", 1)]
        sci = fd[("SCI",1)].data
        fd.close()
        nrows = getNcounts(sci)
        if nrows == 0:
            info["npix"] = (0,)

    hdu = cosutil.createCorrtagHDU(nrows, info["detector"], sci_hdu)
    hdu.header["extname"] = "EVENTS"

    if nrows > 0:
        # Create pseudo-timetag arrays (x & y, no time) from the raw image.
        x = np.zeros(nrows, dtype=np.float32)
        y = np.zeros(nrows, dtype=np.float32)
        ccos.unbinaccum(sci, x, y, x_offset)

        # Copy x and y to the pseudo time-tag table.
        outdata = hdu.data
        outdata.field("TIME")[:] = info["exptime"] / 2.
        outdata.field("RAWX")[:] = x
        outdata.field("RAWY")[:] = y
        outdata.field("XCORR")[:] = x
        outdata.field("YCORR")[:] = y
        outdata.field("XDOPP")[:] = x
        outdata.field("XFULL")[:] = x
        outdata.field("YFULL")[:] = y
        outdata.field("WAVELENGTH")[:] = np.zeros(nrows, dtype=np.float32)
        outdata.field ("EPSILON")[:] = np.ones(nrows, dtype=np.float32)
        outdata.field("DQ")[:] = np.zeros(nrows, dtype=np.int16)
        outdata.field("PHA")[:] = 0

    primary_hdu = fits.PrimaryHDU(header=phdr)
    ofd = fits.HDUList(primary_hdu)
    cosutil.updateFilename(ofd[0].header, outtag)
    ofd.append(hdu)
    ofd.append(cosutil.dummyGTI(info["exptime"]))
    ofd[0].header["nextend"] = len(ofd) - 1     # number of extensions

    ofd.writeto(outtag)
    del ofd

    status = timetag.timetagBasicCalibration(input, inpha, outtag,
                    outflt, outcounts, None, outcsum,
                    cl_args, info, switches, reffiles, wavecal_info)

    return status

def acqImage(input, outflt, outcounts, outcsum, cl_args,
             info, switches, reffiles):
    """Do the calibration for ACQ/IMAGE data."""

    livetimefile = cl_args["livetimefile"]

    fd = fits.open(input, mode="copyonwrite")
    nextend = len(fd) - 1
    nimsets = len(fd) // 3
    phdr = fd[0].header
    phdr["cal_ver"] = info["cal_ver"]
    hdr_list = []

    writePrimaryHDU(outcounts, phdr, nextend)
    writePrimaryHDU(outflt, phdr, nextend)
    if outcsum is not None:
        # we'll add the SCI array for each imset to this array
        csum_array = np.zeros((NUV_Y, NUV_X), dtype=np.float32)

    for imset in range(1, nimsets+1):

        sci_hdr = fd[("SCI",imset)].header
        err_hdr = fd[("ERR",imset)].header
        dq_hdr = fd[("DQ",imset)].header
        hdr_list.append(sci_hdr)

        counts_sci = fd[("SCI",imset)].data
        #
        # Allow subsequent floating point processing
        # astype returns a copy by default, not a view
        flt_sci = counts_sci.astype(np.float64)

        dq_array = cosutil.getInputDQ(input, imset)

        doPhotcorr(info, switches, reffiles["imphttab"], phdr, sci_hdr)

        updateGlobrate(sci_hdr, counts_sci, info["exptime"])

        doDqicorr(info, switches, reffiles, phdr, dq_array)

        doDeadcorr(flt_sci, sci_hdr["exptime"], info, switches, reffiles,
                   phdr, sci_hdr, input, livetimefile)

        if outcsum is not None:
            csum_array += flt_sci

        doFlatcorr(flt_sci, switches, reffiles, phdr)

        (C_rate, errC_rate, E_rate, errE_rate) = makeImages(
                        counts_sci, flt_sci, sci_hdr["exptime"])
        appendImset(outcounts, imset, C_rate, errC_rate, dq_array,
                    sci_hdr, err_hdr, dq_hdr)
        appendImset(outflt, imset, E_rate, errE_rate, dq_array,
                    sci_hdr, err_hdr, dq_hdr)
    doStatflag(switches, outflt, outcounts)

    # Calibration switch keywords have been updated in phdr, but the output
    # primary headers have already been written to disk, so those files need
    # to be reopened in order to update these keywords.
    updateSwitches(phdr, outflt, outcounts)

    if outcsum is not None:
        writeCsum(outcsum, phdr, hdr_list, csum_array,
                  cl_args["raw_csum_coords"],
                  cl_args["binx"], cl_args["biny"],
                  cl_args["compress_csum"],
                  cl_args["compression_parameters"])

    fd.close()
    if not (info["aperture"] in APERTURE_NAMES or
            info["targname"] == "DARK" and
            info["aperture"] in OTHER_APERTURE_NAMES):
        raise BadApertureError("APERTURE = %s is not a valid aperture name." %
                               info["aperture"])

def updateGlobrate(hdr, data, exptime):
    """Update the GLOBRATE keyword in the extension header.

    Parameters
    ----------
    hdr: FITS Header object
        The input events extension header.

    data: array_like
        Data array.

    exptime: float
        Exposure time in seconds.
    """

    if data is None or len(data) == 0 or exptime <= 0.:
        globrate = 0.
    else:
        globrate = data.sum(dtype=np.float64) / exptime

    globrate = round(globrate, 4)
    hdr["globrate"] = globrate

def doPhotcorr(info, switches, imphttab, phdr, hdr):
    """Update photometry parameter keywords for imaging data.

    Parameters
    ----------
    info: dictionary
        Header keywords and values.

    switches: dictionary
        Calibration switches.

    imphttab: str
        The name of the imaging photometric parameters table.

    phdr: FITS Header object
        The primary header, photcorr keyword updated in-place.

    hdr: FITS Header object
        The first extension header, updated in-place.
    """

    if info["obstype"] == "IMAGING" and info["detector"] == "NUV":
        cosutil.printSwitch("PHOTCORR", switches)
        if switches["photcorr"] == "PERFORM":
            # If aperture is invalid, phot.doPhot will use PSA.
            obsmode = "cos,nuv," + info["opt_elem"] + "," + info["aperture"]
            phot.doPhot(imphttab, obsmode, hdr)
            phdr["photcorr"] = "COMPLETE"

def doDqicorr(info, switches, reffiles, phdr, dq_array):
    """Update the DQ array using the DQI table.

    Parameters
    ----------
    info: dictionary
        Header keywords and values.

    switches: dictionary
        Calibration switches.

    reffiles: dictionary
        Reference file names.

    phdr: FITS Header object
        Primary header from input file.

    dq_array: array_like
        DQ array from input file, or an array of zeros; this will be
        modified in-place.
    """

    cosutil.printSwitch("DQICORR", switches)

    if switches["dqicorr"] == "PERFORM":

        cosutil.printRef("BPIXTAB", reffiles)

        # update using imaging wavecal xxx
        minmax_shift_dict = {(0, 1024): [0., 0., 0., 0.]}
        minmax_doppler = (0., 0.)
        doppler_boundary = -10          # anywhere below 0
        cosutil.updateDQArray(info, reffiles, dq_array,
                              minmax_shift_dict,
                              minmax_doppler, doppler_boundary, None)

        phdr["dqicorr"] = "COMPLETE"

def doDeadcorr(flt_sci, exptime, info, switches, reffiles,
               phdr, hdr, input, livetimefile):
    """Correct for deadtime."""

    cosutil.printSwitch("DEADCORR", switches)

    if switches["deadcorr"] == "PERFORM":
        cosutil.printRef("DEADTAB", reffiles)
        (dead_rate, dead_method, livetime) = \
                deadtimeCorrection(flt_sci, exptime, reffiles["deadtab"],
                                   info, input, livetimefile)
        hdr["deadrt"] = dead_rate
        hdr["deadmt"] = dead_method
        hdr["livetm"] = livetime
        if dead_method == "SKIPPED":
            # Deadcorr would already be set to COMPLETE if it was done for
            # the first imset, in which case don't change it.  So COMPLETE
            # means it was done for at least one imset.
            if phdr["deadcorr"] != "COMPLETE":
                phdr["deadcorr"] = "SKIPPED"
        else:
            phdr["deadcorr"] = "COMPLETE"

def deadtimeCorrection(flt_sci, exptime, deadtab, info,
                       input, livetimefile):
    """Determine and apply the livetime factor.

    If there are subarrays, the livetime factor is gotten from the digital
    event counter.  If there are no subarrays, the livetime factor is based
    on the actual count rate.

    Parameters
    ----------
    flt_sci: array_like
        The SCI image array, to be corrected in-place.

    exptime: float
        Exposure time for current imset.

    deadtab: str
        Name of reference table of count rates and livetime factors.

    info: dictionary
        Header keywords and values.

    input: str
        Name of input raw file (for writing to livetimefile).

    livetimefile: str or None
        Name of output text file for livetime factors.

    Returns
    -------
    (dead_rate, dead_method, livetime): tuple of float, str, float
        dead_rate is the count rate used for determining the livetime
        factor.  dead_method is a string that indicates which method was
        used for determining the livetime factor; dead_method will be
        "SKIPPED" if the exposure time is zero.  livetime is the livetime
        factor that was used.
    """

    if exptime <= 0.:
        cosutil.printWarning("Can't do deadcorr, exptime = %.6g." % exptime)
        return (0., "SKIPPED", 1.)

    if livetimefile is None:
        fd = None
    else:
        fd = open(livetimefile, "a")
        fd.write("# %s\n" % (input,))

    ncounts = getNcounts(flt_sci)

    live_info = cosutil.getTable(deadtab,
                                 filter={"segment": info["segment"]},
                                 at_least_one=True)
    obs_rate = live_info.field("obs_rate")
    live_factor = live_info.field("livetime")

    # keyword used to print information
    keyword = "MEVENTS"

    # Output count rate from digital event counter (DEC), and corresponding
    # livetime factor.
    dec_countrate = info["countrate"]
    dec_livetime = cosutil.determineLivetime(dec_countrate,
                                             obs_rate, live_factor)
    actual_countrate = float(ncounts) / exptime
    actual_rate_livetime = cosutil.determineLivetime(actual_countrate,
                                                     obs_rate, live_factor)

    if info["subarray"]:
        livetime_source = "digital event counter (%s)" % keyword
        livetime = dec_livetime
        dead_rate = dec_countrate
        dead_method = "MEVENTS"
    else:
        livetime_source = "actual count rate"
        livetime = actual_rate_livetime
        dead_rate = actual_countrate
        dead_method = "DATA"

    flt_sci /= livetime

    print_details = (cosutil.checkVerbosity(VERY_VERBOSE))      # initial value

    if abs(dec_livetime - actual_rate_livetime) > \
            LIVETIME_CRITERION * actual_rate_livetime:
        cosutil.printWarning("livetime estimates differ.")
        print_details = True

    if print_details:
        cosutil.printMsg("  actual countrate and livetime:  %.6g, %6.4f" % \
                         (actual_countrate, actual_rate_livetime))
        cosutil.printMsg("  countrate and livetime from %s:  %.6g, %6.4f" % \
                         (keyword, dec_countrate, dec_livetime))
        cosutil.printMsg("Livetime %6.4f is based on %s." % \
                         (livetime, livetime_source))

    if fd is not None:
        fd.write("actual countrate and livetime:  %.6g, %6.4f\n" %
                 (actual_countrate, actual_rate_livetime))
        fd.write("countrate and livetime from %s:  %.6g, %6.4f\n" %
                 (keyword, dec_countrate, dec_livetime))
        fd.write("livetime %6.4f is based on %s.\n" % \
                 (livetime, livetime_source))

    if fd is not None:
        fd.close()

    return (dead_rate, dead_method, livetime)

def doFlatcorr(flt_sci, switches, reffiles, phdr):
    """Apply flat field correction.

    Parameters
    ----------
    flt_sci: array_like
        The image array, modified in-place.

    switches: dictionary
        Calibration switches.

    reffiles: dictionary
        Reference file names.

    phdr: FITS Header object
        The input primary header.
    """

    cosutil.printSwitch("FLATCORR", switches)

    if switches["flatcorr"] == "PERFORM":

        cosutil.printRef("FLATFILE", reffiles)

        fd = fits.open(reffiles["flatfile"], mode="copyonwrite")
        hdu = fd[1]
        flat = hdu.data
        fd.close()

        (ny, nx) = flat.shape
        x0 = hdu.header.get("origin_x", 0)
        y0 = hdu.header.get("origin_y", 0)

        flt_sci[int(y0):int(y0+ny),int(x0):int(x0+nx)] /= flat

        phdr["flatcorr"] = "COMPLETE"

def doStatflag(switches, outflt, outcounts):
    """Compute statistics and update keywords.

    Parameters
    ----------
    switches: dictionary
        Calibration switches.

    outflt: str
        Name of the output file for flat-fielded count-rate image.

    outcounts: str
        Name of the output file for count-rate image.
    """

    cosutil.printSwitch("STATFLAG", switches)
    if switches["statflag"] == "PERFORM":
        cosutil.doImageStat(outcounts)
        cosutil.doImageStat(outflt)

def updateSwitches(phdr, outflt, outcounts):
    """Update calibration switch keywords in output primary headers.

    Parameters
    ----------
    phdr: FITS Header object
        The input primary header.

    outflt: str
        Name of the output file for the flat-fielded count-rate image.

    outcounts: str
        Name of the output file for the count-rate image.
    """

    for filename in [outflt, outcounts]:
        fd = fits.open(filename, mode="update")
        for keyword in ["photcorr", "dqicorr", "deadcorr", "flatcorr"]:
            fd[0].header[keyword] = phdr[keyword]
        fd.close()

def makeImages(counts_sci, flt_sci, exptime):
    """Create the count rate and error arrays.

    Parameters
    ----------
    counts_sci: array_like
        The SCI image array, counts.

    flt_sci: array_like
        The SCI image array, after deadcorr and flatcorr.

    exptime: float
        The exposure time.

    Returns
    -------
    (C_rate, errC_rate, E_rate, errE_rate): tuple of floats
        C_rate is the count rate array; errC_rate is the error estimate
        for the count rate array; E_rate is the flat-fielded count rate
        array; errE_rate is the error estimate for the flat fielded
        count-rate array.
    """

    if exptime <= 0:
        cosutil.printWarning(
                "Exposure time is zero, so output files are dummy.")
        C_rate = counts_sci * 0.
        E_rate = C_rate.copy()
        errC_rate = C_rate.copy()
        errE_rate = C_rate.copy()
        return (C_rate, errC_rate, E_rate, errE_rate)

    C_rate = counts_sci / exptime
    E_rate = flt_sci / exptime

    counts_sci_temp = np.where(counts_sci < 0., 0., counts_sci)
    errC_rate = np.sqrt(counts_sci_temp) / exptime
    del counts_sci_temp

    # errC_rate will likely have a number of zero values, so we set those
    # to one before dividing.
    errC_rate_temp = np.where(errC_rate == 0., 1., errC_rate)
    errE_rate = E_rate / errC_rate_temp / exptime
    del errC_rate_temp

    return (C_rate, errC_rate, E_rate, errE_rate)

def writeCsum(outcsum, phdr, hdr_list, csum_array,
              raw_csum_coords,
              binx=None, biny=None,
              compress_csum=False,
              compression_parameters="gzip,-0.1"):
    """Write the "calcos sum" (csum) image.

    Parameters
    ----------
    outcsum: str
        Name of output calcos sum image file.

    phdr: FITS Header object
        Primary header.

    hdr_list: list of FITS Header objects
        List of sci extension headers.

    csum_array: array_like
        Data array for SCI extension.

    raw_csum_coords: boolean
        This only affects the COORDFRM keyword value.

    binx: int or None
        Binning factor in the dispersion direction (or None for
        the default binning).

    biny: int or None
        Binning factor in the cross-dispersion direction (or None
        for the default binning).

    compress_csum: boolean
        Compress the csum image?

    compression_parameters: str
        compressionType and quantizeLevel (separated by a comma) for
        the call to fits.CompImageHDU; compressionType can be "rice",
        "gzip", or "hcompress", and quantizeLevel can be e.g. -0.1,
        which means the floating point values will be scaled to integers
        with spacing that corresponds to 0.1 dn (see the doc string for
        fits.CompImageHDU for more details).
    """

    cosutil.printMsg("writing file %s ..." % outcsum, VERY_VERBOSE)

    primary_hdu = fits.PrimaryHDU(header=phdr)
    fd = fits.HDUList(primary_hdu)
    fd[0].header["nextend"] = 1
    fd[0].header["filetype"] = "CALCOS SUM FILE"
    cosutil.updateFilename(fd[0].header, outcsum)
    if raw_csum_coords:
        fd[0].header["coordfrm"] = "raw"
    else:
        fd[0].header["coordfrm"] = "corrected"

    # used later for updating the COUNTS keyword
    counts = csum_array.sum(dtype=np.float64)

    detector = phdr["detector"]

    shape = csum_array.shape
    if binx is None or binx <= 0:
        binx = NUV_BIN_X
    if biny is None or biny <= 0:
        biny = NUV_BIN_Y
    nx = shape[1] // binx
    ny = shape[0] // biny

    # Copy the exposure time keywords to the output headers.
    hdr = hdr_list[0].copy()
    addExptimeKeywords(hdr_list, hdr)

    if binx > 1 or biny > 1:
        binned_array = np.zeros((ny,nx), dtype=np.float32)
        ccos.bin2d(csum_array, binned_array)
    else:
        binned_array = csum_array

    if compress_csum:
        (compType, quantLevel) = compression_parameters.split(",")
        compType = compType.upper() + "_1"
        quantLevel = float(quantLevel)
        fd.append(fits.CompImageHDU(binned_array, header=hdr, name="SCI",
                                    compressionType=compType,
                                    quantizeLevel=quantLevel))
    else:
        fd.append(fits.ImageHDU(data=binned_array, header=hdr, name="SCI"))
    fd[1].header["BUNIT"] = "count"
    fd[1].header["counts"] = counts
    fd[1].header["nuvbinx"] = binx
    fd[1].header["nuvbiny"] = biny

    fd.writeto(outcsum, output_verify="silentfix")

def getNcounts(sci):
    """Return the total number of counts in an array.

    Parameters
    ----------
    sci: array_like
        Image data array, in counts.

    Returns
    -------
    ncounts: int
        Sum of values in sci, rounded to an int.
    """

    ncounts = sci.sum(dtype=np.float64)
    # The value returned by sum() is an "array scalar," so convert it.
    ncounts = float(ncounts)
    ncounts = int(round(ncounts))

    return ncounts

def addExptimeKeywords(hdr_list, hdr):
    """Copy the exposure time keywords to the output headers.

    exptime and rawtime will be the sums of those values from all (both)
    of the input headers, and expstart and expend will be taken from
    the first and last headers.

    Parameters
    ----------
    hdr_list: list of FITS Header objects
        List of sci extension headers.

    hdr: FITS Header object
        Output header, modified in-place
    """

    # The headers in hdr_list are assumed to be in chronological order.
    expstart = hdr_list[0].get("expstart", -999.)
    expend = hdr_list[-1].get("expend", -999.)

    exptime = 0.
    rawtime = 0.
    for hdr_i in hdr_list:
        exptime_i = hdr_i.get("exptime", 0.)
        exptime += exptime_i
        rawtime += hdr_i.get("rawtime", exptime_i)

    hdr["expstart"] = expstart
    hdr["expend"] = expend
    hdr["exptime"] = exptime
    hdr["rawtime"] = rawtime

def writePrimaryHDU(output, phdr, nextend):
    """Write an output file containing just a primary header.

    Parameters
    ----------
    output: str
        Name of the output file (flt or counts).

    phdr: FITS Header object
        Primary header of input file.

    nextend: int
        Number of extensions.
    """

    cosutil.printMsg("writing file %s ..." % output, VERY_VERBOSE)

    primary_hdu = fits.PrimaryHDU(header=phdr)
    fd = fits.HDUList(primary_hdu)
    fd[0].header["nextend"] = nextend
    cosutil.updateFilename(fd[0].header, output)

    fd.writeto(output, output_verify="silentfix")

def appendImset(output, imset, sci_array, err_array, dq_array,
                sci_hdr, err_hdr, dq_hdr):
    """Append an image set (SCI, ERR, DQ extensions).

    This function appends one image set to an output file.

    Parameters
    ----------
    output: str
        Name of the output file (flt or counts).

    imset: int
        Image set number (one indexed, to match EXTVER).

    sci_array: array_like or None
        Data array for the SCI extension.

    err_array: array_like or None
        Data array for the ERR extension.

    dq_array: array_like or None
        Data array for the DQ extension.

    sci_hdr: FITS Header object
        Header for SCI extension.

    err_hdr: FITS Header object
        Header for ERR extension.

    dq_hdr: FITS Header object
        Header for DQ extension.
    """

    fd = fits.open(output, mode="update")

    hdu = fits.ImageHDU(data=sci_array, header=sci_hdr, name="SCI")
    hdu.header["EXTVER"] = imset
    hdu.header["BUNIT"] = "count /s"
    fd.append(hdu)

    hdu = fits.ImageHDU(data=err_array, header=err_hdr, name="ERR")
    hdu.header["EXTVER"] = imset
    hdu.header["BUNIT"] = "count /s"
    fd.append(hdu)

    hdu = fits.ImageHDU(data=dq_array, header=dq_hdr, name="DQ")
    hdu.header["EXTVER"] = imset
    hdu.header["BUNIT"] = "UNITLESS"
    fd.append(hdu)

    fd.close()



import math
import numpy as np
from astropy.io import fits

TWOPI       = 2. * math.pi
SEC_PER_DAY = 86400.0

class HSTOrbit(object):
    """Orbital parameters.

    The public methods are getOrbitper and getPos.

    This was originially written in IDL (hst_pos_mjd.pro) by Tom Ake.

    Parameters
    ----------
    sptfile: str
        The name of the support file (rootname_spt.fits).
    """

    def __init__(self, sptfile):
        """Constructor."""

        # attributes
        self.argperig = 0.  # argument of perigee (revolutions)
        self.cirveloc = 0.  # circular orbit linear velocity (meters/second)
        self.cosincli = 0.  # cosine of inclination
        self.ecbdx3   = 0.  # eccentricity cubed times 3
        self.eccentry = 0.  # eccentricity
        self.eccentx2 = 0.  # eccentricity times 2
        self.ecbdx4d3 = 0.  # eccentricity cubed times 4/3
        self.epchtime = 0.  # epoch time of parameters (secs since 1/1/85)
        self.esqdx5d2 = 0.  # eccentricity squared times 5/2
        self.fdmeanan = 0.  # 1st derivative coef for mean anomly (revs/sec)
        self.hsthorb  = 0.  # half the duration of the ST orbit (seconds)
        self.meananom = 0.  # mean anomaly (radians)
        self.rascascn = 0.  # right ascension of ascending node (revolutions)
        self.rcargper = 0.  # rate change of argument of perigee (revs/sec)
        self.rcascnrv = 0.  # rt chge right ascension ascend node (revs/sec)
        self.sdmeanan = 0.  # 2nd deriv coef for mean anomaly (revs/sec/sec)
        self.semilrec = 0.  # semi-latus rectum (meters)
        self.sineincl = 0.  # sine of inclination

        self._readOrbitalParameters(sptfile)

    def _readOrbitalParameters(self, sptfile):
        """Get the orbital parameters from the spt primary header.

        Parameters
        ----------
        sptfile: str
            The name of the support file.
        """

        fd = fits.open(sptfile, mode="readonly")
        phdr = fd[0].header

        # Orbital elements for HST.
        self.argperig = phdr["argperig"]
        self.cirveloc = phdr["cirveloc"]
        self.cosincli = phdr["cosincli"]
        self.ecbdx3   = phdr["ecbdx3"]
        self.eccentry = phdr["eccentry"]
        self.eccentx2 = phdr["eccentx2"]
        self.ecbdx4d3 = phdr["ecbdx4d3"]
        self.epchtime = phdr["epchtime"]
        self.esqdx5d2 = phdr["esqdx5d2"]
        self.fdmeanan = phdr["fdmeanan"]
        self.hsthorb  = phdr["hsthorb"]
        self.meananom = phdr["meananom"]
        self.rascascn = phdr["rascascn"]
        self.rcargper = phdr["rcargper"]
        self.rcascnrv = phdr["rcascnrv"]
        self.sdmeanan = phdr["sdmeanan"]
        self.semilrec = phdr["semilrec"]
        self.sineincl = phdr["sineincl"]

        fd.close()

    def getOrbitper(self):
        """Return the orbital period.

        Returns
        -------
        float
            The orbital period in seconds.
        """

        return 2. * self.hsthorb

    def getPos(self, mjd):
        """Get position and velocity at a given time.

        # S. Hulbert, Oct 91    Original
        # PEH, 2008 Oct 3       Converted from SPP to Python

        Parameters
        ----------
        mjd: float
            The time (MJD) at which to compute the position and velocity.

        Returns
        -------
        tuple of two array_like
            The first array is the position vector (km), the second array
            is the velocity vector (km/s).
        """

        # These will be returned, after assigning the actual values.
        x_hst = np.zeros(3, dtype=np.float64)
        v_hst = np.zeros(3, dtype=np.float64)

        argperig = self.argperig
        cirveloc = self.cirveloc
        cosincli = self.cosincli
        ecbdx3   = self.ecbdx3
        eccentry = self.eccentry
        eccentx2 = self.eccentx2
        ecbdx4d3 = self.ecbdx4d3
        epchtime = self.epchtime
        esqdx5d2 = self.esqdx5d2
        fdmeanan = self.fdmeanan
        hsthorb  = self.hsthorb
        meananom = self.meananom
        rascascn = self.rascascn
        rcargper = self.rcargper
        rcascnrv = self.rcascnrv
        sdmeanan = self.sdmeanan
        semilrec = self.semilrec
        sineincl = self.sineincl

        # convert time from MJD to seconds since 1985 Jan 1
        sec85 = (mjd - 46066.0) * SEC_PER_DAY

        # calculate time difference between observation and epoch time
        deltim = sec85 - epchtime

        # mean anomaly
        temp2 = fdmeanan * deltim
        temp3 = 0.5 * sdmeanan * deltim*deltim
        m = meananom + TWOPI * (temp2 + temp3)

        sin_m = math.sin(m)
        cos_m = math.cos(m)

        # true anomaly (equation of the center)
        v = m + sin_m * (eccentx2 + ecbdx3 * cos_m * cos_m -
                ecbdx4d3 * sin_m * sin_m + esqdx5d2 * cos_m)
        sin_v = math.sin(v)
        cos_v = math.cos(v)

        # distance
        r = semilrec / (1.0 + eccentry * cos_v)

        # argument of perigee
        wsmall = TWOPI * (argperig + rcargper * deltim)

        # longitude of the ascending node
        wbig = TWOPI * (rascascn + rcascnrv * deltim)
        sin_wbig = math.sin(wbig)
        cos_wbig = math.cos(wbig)

        # calculate the rectangular coordinates
        #  (see Smart, Spherical Astronomy, section 75, page 122-124)

        f = wsmall + v
        sin_f = math.sin(f)
        cos_f = math.cos(f)

        x_hst[0] = r * (cos_wbig * cos_f - cosincli * sin_wbig * sin_f)
        x_hst[1] = r * (sin_wbig * cos_f + cosincli * cos_wbig * sin_f)
        x_hst[2] = r * sineincl * sin_f

        a0 = cirveloc * eccentry * sin_v / r
        a1 = cirveloc * (1.0 + eccentry * cos_v) + \
                TWOPI * rcargper * r
        v_hst[0] = a0 * x_hst[0] - \
                a1 * (cos_wbig * sin_f + cosincli * sin_wbig * cos_f) - \
                TWOPI * rcascnrv * x_hst[1]
        v_hst[1] = a0 * x_hst[1] - \
                a1 * (sin_wbig * sin_f - cosincli * cos_wbig * cos_f) + \
                TWOPI * rcascnrv * x_hst[0]
        v_hst[2] = a0 * x_hst[2] + a1 * sineincl * cos_f

        # Convert from meters to kilometers.
        x_hst /= 1000.0
        v_hst /= 1000.0

        return (x_hst, v_hst)



import io
import os
import sys
import time
from pathlib import Path

import numpy as np
import pytest
from astropy.io import fits

from calcos import cosutil
from calcos.calcosparam import MissingRowError
from generate_tempfiles import generate_fits_file


def test_find_column(tmp_path):
    # Setup
    # create a test fits file
    name = str(tmp_path / "findCol.fits")
    ofd = generate_fits_file(name)

    target_col = 'TIME'
    # Test
    col_exists = True
    # Verify
    assert col_exists == cosutil.findColumn(name, target_col)


def test_get_table(tmp_path):
    # Setup
    # create a test fits file
    name = str(tmp_path / "getTable.fits")
    ofd = generate_fits_file(name)
    truth = [tuple(ofd[1].data[3])]
    time = ofd[1].data[3][0]
    # Test
    dt = list(cosutil.getTable(name, {'TIME': time}, exactly_one=True))
    # Verify
    np.testing.assert_array_equal(truth, dt)


def test_get_table_exceptions(tmp_path):
    # Raise MissingRowError
    name = str(tmp_path / "getTable.fits")
    generate_fits_file(name)
    # truth = [tuple(ofd[1].data[3])]
    t = 1.0  # non-existent value
    with pytest.raises(MissingRowError):
        cosutil.getTable(name, {'Time': t}, exactly_one=True)


def test_get_col_copy(tmp_path):
    # Setup
    # create a test fits file
    name = str(tmp_path / "getTable.fits")
    ofd = generate_fits_file(name)
    col_name = 'XCORR'
    portion_of_array = ofd[1].data[:]
    truth_values = ofd[1].data.field(col_name)
    # Test (using filename or portion of data)
    test1 = cosutil.getColCopy(name, column=col_name)
    test2 = cosutil.getColCopy(column=col_name, data=portion_of_array)

    # Verify
    np.testing.assert_array_equal(truth_values, test1)
    np.testing.assert_array_equal(truth_values, test2)


def test_get_col_copy_exception(tmp_path):
    # raise RuntimeError error
    with pytest.raises(RuntimeError):
        name = str(tmp_path / "getTable.fits")
        ofd = generate_fits_file(name)
        col_name = 'XCORR'
        portion_of_array = ofd[1].data[:]
        cosutil.getColCopy(filename=str(tmp_path / "Output/getTable.fits"), column=col_name, data=portion_of_array)
        cosutil.getColCopy(filename=None, column=col_name, data=None)


def test_get_headers(tmp_path):
    # Setup
    # create a test fits file
    name = str(tmp_path / "getHeaders.fits")
    ofd = generate_fits_file(name)
    true_hdr = ofd[0].header

    # Test
    test_hdr = cosutil.getHeaders(name)

    # Verify
    np.testing.assert_array_equal(true_hdr, test_hdr[0])


def test_write_output_events(tmp_path):
    # Setup
    in_file = str(tmp_path / "outputEvents.fits")
    out_file = str(tmp_path / "outputEvents_cpy.fits")
    generate_fits_file(in_file)
    actual_lines = 10
    lines = cosutil.writeOutputEvents(in_file, out_file)
    assert actual_lines == lines


def test_concat_arrays():
    # setup
    arr1 = np.ones(10, dtype=float)
    arr2 = np.zeros(10, dtype=float)
    actual = np.concatenate((arr1, arr2), axis=0)
    # Test
    concat_arrays = cosutil.concatArrays(arr1, arr2)
    # Verify
    np.testing.assert_array_equal(actual, concat_arrays)


def test_update_filename(tmp_path):
    # Setup
    path = tmp_path / "update_filename.fits"
    filename = str(path)
    generate_fits_file(filename)
    before_update_hdr = fits.open(filename, mode="update")
    # Test
    cosutil.updateFilename(before_update_hdr[0].header, path.stem)
    before_update_hdr.close()
    after_update_hdr = fits.open(filename)
    # Verity
    assert path.stem == after_update_hdr[0].header["filename"]
    after_update_hdr.close()


def test_copy_file(tmp_path):
    # Setup
    infile = str(tmp_path / "input.fits")
    generate_fits_file(infile)
    outfile = str(tmp_path / "output.fits")
    # Test
    cosutil.copyFile(infile, outfile)
    # Verify
    inf = fits.open(infile)
    out = fits.open(outfile)
    np.testing.assert_array_equal(inf[1].data, out[1].data)
    np.testing.assert_array_equal(inf[2].data, out[2].data)
    np.testing.assert_array_equal(inf[3].data, out[3].data)


def test_is_product(tmp_path):
    # Setup
    product_file = str(tmp_path / "my0_product_a.fits")
    raw_file = str(tmp_path / "my_raw.fits")
    generate_fits_file(product_file)
    generate_fits_file(raw_file)
    # Test

    # NOTE:
    # no test to be done here since we're checking the file if its a product or not
    # the return of the function isProduct() is a boolean hence, assert it directly.

    # Verify
    assert cosutil.isProduct(product_file)
    assert not cosutil.isProduct(raw_file)




def test_cmp_part_exception():
    # test for exception
    with pytest.raises(TypeError) as t_err:
        cosutil.cmpPart(5, "k")


def test_cmp_part():
    # Setup
    str1 = "hello"
    str2 = "hey"
    # Test
    cmp1 = cosutil.cmpPart(str1, str2)
    cmp2 = cosutil.cmpPart(str2, str1)
    cmp3 = cosutil.cmpPart(str1, str1)
    # Verify
    assert cmp1 == -1
    assert cmp2 == 1
    assert cmp3 == 0


def test_split_int_letter():
    # Setup
    test = ["1this", "is", "4a", "15test"]
    truth = ["1", "this", "0", "is", "4", "a", "15", "test"]
    # Test
    cosutil.splitIntLetter(test)
    # Verify
    assert truth == test


def test_create_corrtag_hdu(tmp_path):
    # Setup
    hdu = generate_fits_file(str(tmp_path / "corrtag.fits"))
    num_of_rows = 10
    # Test
    # detector parameter is not needed consider removing it
    out_bin_table = cosutil.createCorrtagHDU(num_of_rows, detector="FUV", hdu=hdu[0])
    assert len(out_bin_table.data) == num_of_rows
    assert all(out_bin_table.header) == all(hdu[0].header)


def test_remove_wcs_keywords(tmp_path):
    # Setup
    hdu = generate_fits_file(str(tmp_path / "removeWCS.fits"))
    hdu2 = generate_fits_file(str(tmp_path / "corrtag.fits"))
    inhdr = hdu2[1].header
    cd = hdu[1].data.columns
    WCS_keywords = ['TCTYP*',
                    'TCUNI*',
                    'TCRPX*',
                    'TCRVL*',
                    'TCDLT*',
                    ]
    # Test
    newheader = cosutil.remove_WCS_keywords(inhdr, cd)
    # Verify
    for keys in WCS_keywords:
        assert keys not in newheader
        assert len(inhdr[keys]) > 0

def test_dummy_gti():
    # Setup
    test_exptime_value = 1.423
    # Test
    dummy_hdu = cosutil.dummyGTI(test_exptime_value)
    # Verify
    assert dummy_hdu.data[0][0] == 0.0
    assert dummy_hdu.data[0][1] == test_exptime_value


def test_return_gti(tmp_path):
    filename = str(tmp_path / "gti_file.fits")
    # Setup
    hdu = generate_fits_file(filename)
    # Test
    gti = cosutil.returnGTI(filename)
    # Verify
    np.testing.assert_array_equal(list(hdu[2].data), gti)


def test_err_frequentist():
    """
    unit test for err_frequentist(counts)
    - create 3 arrays.
    - find the poisson confidence interval for each array.
    - assert the result with the expected err_lower and err_upper.
    """
    # Setup
    zeros = np.zeros(5)
    ones = np.ones(5)
    random_values = np.array([2.2400559, 0.85776844, 5.31731382, 8.98167105, 7.88191824]).astype(np.float32)

    # Actual values expected
    true_lower1 = np.zeros(5).astype(np.float32)
    true_upper1 = np.array([1.8410217, 1.8410217, 1.8410217, 1.8410217, 1.8410217]).astype(np.float32)

    true_lower2 = np.array([0.82724625, 0.82724625, 0.82724625, 0.82724625, 0.82724625]).astype(np.float32)
    true_upper2 = np.array([2.2995265, 2.2995265, 2.2995265, 2.2995265, 2.2995265]).astype(np.float32)

    true_lower3 = np.array([1.3812032, 0.74085414, 2.2319276, 2.940346, 2.7469769]).astype(np.float32)
    true_upper3 = np.array([2.7093022, 2.2441676, 3.4480913, 4.1072574, 3.9250364]).astype(np.float32)
    # Test
    lower1, upper1 = cosutil.errFrequentist(zeros)
    lower2, upper2 = cosutil.errFrequentist(ones)
    lower3, upper3 = cosutil.errFrequentist(random_values)

    # Verify
    np.testing.assert_array_equal(true_lower1, lower1)
    np.testing.assert_array_equal(true_upper1, upper1)
    np.testing.assert_array_equal(true_lower2, lower2)
    np.testing.assert_array_equal(true_upper2, upper2)
    np.testing.assert_array_equal(true_lower3, lower3)
    np.testing.assert_array_equal(true_upper3, upper3)


def test_precess():
    """
    unit test for precess(t, target)
    - set a time in MJD
    - create a unit vector toward the target
    - calculate the expected coordinates
    - assert expected with the actual.
    """
    # Setup
    time = 55545.270617  # MJD
    target = np.array([3.4, 6.5, 2.5])
    actual_coordinates = [3.3814054391781228, 6.508305447528035, 2.5036088867020565]
    # Test
    test_coordinates = cosutil.precess(time, target)
    # Verify
    np.testing.assert_array_equal(actual_coordinates, test_coordinates)


def test_fit_quartic():
    # Setup
    x = np.array([1, 2, 3, 4, 5, 6])
    y = np.array([2, 4, 6, 8, 10, 12])
    # Expected
    fitted_polynomial = ([-2.00000000e+00, 2.00000000e+00, 8.46875656e-12, -1.74897134e-12,
                          1.23333442e-13],
                         [4.97529008e-11, 1.33375950e-10, 3.65406961e-11, 1.56942681e-12,
                          7.95573614e-15])
    test_polynomial = cosutil.fitQuartic(x, y)
    # Verify
    np.testing.assert_almost_equal(fitted_polynomial, test_polynomial)


def test_center_of_quadratic():
    """
    unit test for center_of_quadratic(coeff, var)
    - create a randomized coeff and var arrays
    - follow the math to calculate x_min and x_min_sigma aka center of the quadratic
    - x_min = -coeff[1] / (2 * coeff[2])
    - x_min_sigma = 0.5 * math.sqrt(var1 / coeff[2] ** 2 + var2 * coeff[1] ** 2 / coeff[2] ** 4)
    - assert the expected result with the functions return.
    """
    # Setup
    coeff1 = np.array([2, 4, 1])
    coeff2 = np.array([3, 5, 7])
    coeff3 = np.array([2.3, 4.6, 1.3])
    coeff4 = np.zeros(shape=3)

    var1 = np.array([0, 2, 4])
    var2 = np.array([1, 2, 5])
    var3 = np.array([5, 3, 9])
    # Expected
    actual_center1 = (-2.0, 4.06201920231798)
    actual_center2 = (-0.35714285714285715, 0.15237943390885794)
    actual_center3 = (-1.769230769230769, 4.1368310795286165)
    actual_center4 = (None, 0.)
    # Test
    center1 = cosutil.centerOfQuadratic(coeff1, var1)
    center2 = cosutil.centerOfQuadratic(coeff2, var2)
    center3 = cosutil.centerOfQuadratic(coeff3, var3)
    center4 = cosutil.centerOfQuadratic(coeff4, var3)
    # Verify
    assert actual_center1 == center1
    assert actual_center2 == center2
    assert actual_center3 == center3
    assert actual_center4 == center4


def test_fit_quadratic():
    # Setup
    x = np.array([1, 2, 4])
    y = np.array([0, 7, 15])
    # Expected
    expected_quadratic = (np.array([-9., 10., -1.], dtype=np.float64), np.array([0., 0., 0.]))
    # Test
    fitted_quadratic = cosutil.fitQuadratic(x, y)
    # round to 10 decimal places to avoid assertion error due to floating point error.
    fitted_quadratic[0][0] = round(fitted_quadratic[0][0], 10)
    fitted_quadratic[0][1] = round(fitted_quadratic[0][1], 10)
    fitted_quadratic[0][2] = round(fitted_quadratic[0][2], 10)
    # Verify
    np.testing.assert_array_equal(expected_quadratic[0], fitted_quadratic[0])
    np.testing.assert_array_equal(expected_quadratic[1], fitted_quadratic[1])


def test_change_segment(tmp_path):
    # Setup
    filename1 = str(tmp_path / "testfits_a.fits")
    filename2 = str(tmp_path / "testfits_b.fits")
    filename3 = str(tmp_path / "testfits.fits")
    # Expected
    fname1 = filename2
    fname2 = filename1
    fname3 = filename3
    # Test
    test_name1 = cosutil.changeSegment(filename1, "FUV", "FUVB")
    test_name2 = cosutil.changeSegment(filename2, "FUV", "FUVA")
    test_name3 = cosutil.changeSegment(filename3, "NUV", "")
    # Verify
    assert fname1 == test_name1
    assert fname2 == test_name2
    assert fname3 == test_name3


def test_copy_exptime_keywords(tmp_path):
    # Setup
    # create two files
    files = [str(tmp_path / "original.fits"), str(tmp_path / "copy.fits")]
    for file in files:
        generate_fits_file(file)
    headers = ["expstart", "expend", "exptime", "rawtime"]
    # set values to the exposure time
    for file in files:
        with open(file, "ab+"):
            for header in headers:
                if file == files[0]:
                    fits.setval(file, header, value=-999, ext=1)
                # set values in the copy to 0 to check if the function is actually changing the values.
                else:
                    fits.setval(file, header, value=0, ext=1)
    # get header of the files
    inhdr = fits.getheader(files[0], 1)
    outhdr = fits.getheader(files[1], 1)
    # Test
    cosutil.copyExptimeKeywords(inhdr, outhdr)
    # Verify
    for header in headers:
        assert inhdr[header] == outhdr[header]

def test_copy_voltage_keywords(tmp_path):
    # Setup
    original = [str(tmp_path / "originalFUV.fits"), str(tmp_path / "originalNUV.fits")]
    copy = [str(tmp_path / "copyFUV.fits"), str(tmp_path / "copyNUV.fits")]

    # create two files each for FUV and NUV
    for file in original:
        generate_fits_file(file)
    for file in copy:
        generate_fits_file(file)
    detectors = ["FUV", "NUV"]
    fuv_headers = ["dethvla", "dethvlb", "dethvca", "dethvcb", "dethvna", "dethvnb"]
    nuv_headers = ["dethvl", "dethvc"]
    # set the values to the headers
    index = 0
    for fileName in original:
        detector = detectors[index]
        with open(fileName, "ab+"):
            if detector == "FUV":
                for header in fuv_headers:
                    fits.setval(fileName, header, value=-999., ext=1)
            else:
                for header in nuv_headers:
                    fits.setval(fileName, header, value=-999, ext=1)
        index += 1
    index = 0
    for fileName in copy:
        detector = detectors[index]
        with open(fileName, "ab+"):
            if detector == "FUV":
                for header in fuv_headers:
                    fits.setval(fileName, header, value=0., ext=1)
            else:
                for header in nuv_headers:
                    fits.setval(fileName, header, value=0., ext=1)
        index += 1
    in_FUV_hdr = fits.getheader(original[0], 1)
    in_NUV_hdr = fits.getheader(original[1], 1)
    out_FUV_hdr = fits.getheader(copy[0], 1)
    out_NUV_hdr = fits.getheader(copy[1], 1)
    # Test 1
    cosutil.copyVoltageKeywords(in_FUV_hdr, out_FUV_hdr, detectors[0])
    # Verify 1
    for header in fuv_headers:
        assert in_FUV_hdr[header] == out_FUV_hdr[header]
    # Test 2
    cosutil.copyVoltageKeywords(in_NUV_hdr, out_NUV_hdr, detectors[1])
    # Verify 2
    for header in nuv_headers:
        assert in_NUV_hdr[header] == out_NUV_hdr[header]

def test_copy_sub_keywords(tmp_path):
    # Setup
    files = [str(tmp_path / "subKeywords.fits"), str(tmp_path / "copySubKeywords.fits")]
    for file in files:
        generate_fits_file(file)
    headers = ["corner%1dx", "corner%1dy", "size%1dx", "size%1dy"]

    for file in files:
        with open(file, "ab+"):
            for header in headers:
                for i in range(8):
                    if file == files[0]:
                        fits.setval(file, header % i, value=-1, ext=1)
                    else:  # set the value of the copy to 0 to test for change.
                        fits.setval(file, header % i, value=0, ext=1)
    with open(files[0], "ab+"):
        fits.setval(files[0], "nsubarry", value=0, ext=1)
    with open(files[1], "ab+"):
        fits.setval(files[1], "nsubarry", value=-1, ext=1)
    inhdr = fits.getheader(files[0], 1)
    outhdr = fits.getheader(files[1], 1)
    # Test
    cosutil.copySubKeywords(inhdr, outhdr, False)
    # Verify
    for header in headers:
        for i in range(8):
            assert inhdr[header % i] == outhdr[header % i]
    cosutil.copySubKeywords(inhdr, outhdr, True)
    # check if nsubarry has been set to 0
    assert inhdr["nsubarry"] == outhdr["nsubarry"]


def test_modify_asn_mtyp():
    # Setup
    str1 = "EXP-TESTVALUE"
    str2 = "EXP_SECONDVALUE"
    str3 = "VALUE_WITHOUT_PREFIX"
    # Expected vals
    string1 = "PROD-TESTVALUE"
    string2 = "PROD_SECONDVALUE"
    string3 = str3
    # Test
    val1 = cosutil.modifyAsnMtyp(str1)
    val2 = cosutil.modifyAsnMtyp(str2)
    val3 = cosutil.modifyAsnMtyp(str3)
    # Verify
    assert string1 == val1
    assert string2 == val2
    assert string3 == val3


def test_rename_file(tmp_path):
    # Setup
    original_filename1 = str(tmp_path / "raw-file.fits")
    original_filename2 = str(tmp_path / "product0_file_a.fits")

    new_filename1 = str(tmp_path / "renamed_raw_file.fits")
    new_filename2 = str(tmp_path / "renamed0_file_a.fits")
    # Create the files
    generate_fits_file(original_filename1)
    generate_fits_file(original_filename2)
    # Test
    cosutil.renameFile(original_filename1, new_filename1)
    cosutil.renameFile(original_filename2, new_filename2)
    # Verify
    assert os.path.exists(new_filename1)
    assert os.path.exists(new_filename2)


def test_del_corrtag_wcs(tmp_path):
    # Setup
    filename = str(tmp_path / "del_corrtagWCS.fits")
    generate_fits_file(filename)
    thdr = fits.getheader(filename, 3)
    tkey = ["TCTYP2", "TCRVL2", "TCRPX2", "TCDLT2", "TCUNI2", "TC2_2", "TC2_3",
            "TCTYP3", "TCRVL3", "TCRPX3", "TCDLT3", "TCUNI3", "TC3_2", "TC3_3"]
    # Test
    thdr = cosutil.delCorrtagWCS(thdr)
    # Verify
    for key in tkey:
        assert key not in thdr


def test_set_verbosity():
    # Setup
    verbosity = 5
    # Test
    cosutil.setVerbosity(verbosity)
    test_value = cosutil.verbosity
    # Verify
    assert verbosity == test_value


def test_check_verbosity():
    # Setup
    verbosity = 4
    # Test
    cosutil.setVerbosity(3)
    check_verbosity = cosutil.checkVerbosity(verbosity)  # should be False
    cosutil.setVerbosity(verbosity)
    check2 = cosutil.checkVerbosity(verbosity)  # should be True
    # Verify
    assert not check_verbosity
    assert check2


def test_set_write_to_trailer():
    # Setup
    flag = True
    # Test
    cosutil.setWriteToTrailer(flag)  # should change to True
    flag_value = cosutil.write_to_trailer
    cosutil.setWriteToTrailer()  # should change to False
    # Verify
    assert flag_value
    assert not cosutil.write_to_trailer


def test_print_msg():
    # Setup
    captured_msg = io.StringIO()
    sys.stdout = captured_msg  # redirect stdout
    verbosity_level = 4
    test_message = "message from pytest"
    # Test
    cosutil.setVerbosity(verbosity_level)
    cosutil.printMsg(test_message, verbosity_level)
    sys.stdout = sys.__stdout__  # reset the redirect
    # Verify
    assert test_message == captured_msg.getvalue()[:-1]  # to remove the newline at the end


def test_return_time():
    t = time.strftime("%d-%b-%Y %H:%M:%S %Z", time.localtime(time.time()))
    get_time = cosutil.returnTime()
    assert t == get_time


"""
Unit tests that have the word "print" in their name using the same algorithm.
- open an IO stream
- initialize the message you want to print
- call the function that is being tested and pass the message string to it.
- redirect the output stream towards the function to catch the printed message
- write the value to a variable
- assert the captured message with the original message.
"""


def test_print_intro():
    # Setup
    captured_msg = io.StringIO()
    sys.stdout = captured_msg
    verbosity = 4
    message = "Pytest message"
    # Test
    cosutil.setVerbosity(verbosity)
    cosutil.printIntro(message)
    sys.stdout = sys.__stdout__
    # remove the newlines at the beginning and end
    captured_msg = captured_msg.getvalue()[1:-1]
    time = cosutil.returnTime()
    message = message + " -- " + time
    # Verify
    assert captured_msg == message


def test_print_filenames():
    # Setup
    captured_msg = io.StringIO()
    sys.stdout = captured_msg
    verbosity = 4
    names = [("Input", "abc_raw.fits"), ("Output", "abc_flt.fits")]
    stimfile = "stim.txt"
    livetimefile = "live.txt"
    expected_msg = ""
    for (lable, filename) in names:
        expected_msg += "%-10s%s\n" % (lable, filename)
    expected_msg += "stim locations log file   " + stimfile + "\n"
    expected_msg += "livetime factors log file " + livetimefile + "\n"
    # Test
    cosutil.setVerbosity(verbosity)
    cosutil.printFilenames(names, shift_file=None, stimfile=stimfile, livetimefile=livetimefile)
    sys.stdout = sys.__stdout__
    # Verify
    assert expected_msg == captured_msg.getvalue()


def test_print_warning():
    # Setup
    captured_msg = io.StringIO()
    sys.stdout = captured_msg
    verbosity = 4
    message = "warning message from pytest"
    # Test
    cosutil.setVerbosity(verbosity)
    cosutil.printWarning(message, verbosity)
    sys.stdout = sys.__stdout__
    message = "Warning:  " + message + "\n"
    # Verify
    assert message == captured_msg.getvalue()


def test_print_error():
    # Setup
    captured_msg = io.StringIO()
    sys.stdout = captured_msg
    verbosity = 4
    message = "error message from pytest"
    # Test
    cosutil.setVerbosity(verbosity)
    cosutil.printError(message)
    sys.stdout = sys.__stdout__
    message = "ERROR:  " + message + "\n"
    # Verify
    assert message == captured_msg.getvalue()


def test_print_continuation():
    # Setup
    captured_msg = io.StringIO()
    sys.stdout = captured_msg
    verbosity = 4
    message = "continued message from pytest"
    # Test
    cosutil.setVerbosity(verbosity)
    cosutil.printContinuation(message, level=verbosity)
    sys.stdout = sys.__stdout__
    message = "    " + message + "\n"
    assert message == captured_msg.getvalue()


def test_print_ref():
    # Setup
    captured_msg = io.StringIO()
    sys.stdout = captured_msg
    verbosity = 4
    keyw = "flatfile".upper()
    reffiles = {"flatfile": "abc_flat.fits", "flatfile_hdr": "lref$abc_flat.fits"}
    message = "%-8s= %s" % (keyw, reffiles[keyw.lower() + "_hdr"]) + "\n"
    cosutil.setVerbosity(verbosity)
    # Test
    cosutil.printRef(keyw, reffiles)
    sys.stdout = sys.__stdout__
    # Verify
    assert message == captured_msg.getvalue()


def test_print_switch():
    # Setup
    switches = {"statflag": "PERFORM", "flatcorr": "PERFORM", "geocorr": "COMPLETE", "randcorr": "SKIPPED"}
    keys = ["STATFLAG", "FLATCORR", "GEOCORR", "RANDCORR"]
    captured_msg = io.StringIO()
    sys.stdout = captured_msg
    msg1 = "STATFLAG  T\n"
    msg2 = "FLATCORR  PERFORM\n"
    msg3 = "%-9s OMIT (already complete)" % keys[2].upper() + "\n"
    msg4 = "%-9s OMIT (skipped)" % keys[3].upper() + "\n"
    actual_msg = msg1 + msg2 + msg3 + msg4
    # Test
    for key in keys:
        cosutil.printSwitch(key, switches)
        captured_msg.flush()
    sys.stdout = sys.__stdout__
    # Verify
    assert actual_msg == captured_msg.getvalue()


def test_guess_aper_from_locn():
    """
    unit test for guessAperFromLocn()
    - create lists for LPs and aperture positions (2 in this case).
    - use the ranges provided to guess which aperture is being used
        LP: 1
        (116.0, 135) ---> PSA
        (-163.0, -143.0) ---> BOA
        LP: 2
        (52.0, 72.0) ---> PSA
        (-227.0, -207.0) ---> BOA
        LP: 3 and above
        aperture will be none
    - assert expected positions with the actual position.
    """
    # Setup
    lps = [1, 2, 3]
    aper_pos1 = [120.2, -151.34, 2.34]  # lp 1
    expected_aper1 = ["PSA", "BOA", None]
    aper_pos2 = [63.2, -214.89, 10.67]  # lp 2
    expected_aper2 = ['PSA', 'BOA', None]
    # Test
    test_postitions1 = []
    test_postitions2 = []
    test_postitions3 = []
    for lp in lps:
        if lp == 1:
            for pos in aper_pos1:
                test_postitions1.append(cosutil.guessAperFromLocn(lp, pos))
        elif lp == 2:
            for pos in aper_pos2:
                test_postitions2.append(cosutil.guessAperFromLocn(lp, pos))
        else:
            test_postitions3 = None
    # Verify
    assert expected_aper1 == test_postitions1
    assert expected_aper2 == test_postitions2
    assert test_postitions3 is None


def test_segment_specific_keyword():
    # Setup
    root = "keyword_root_"
    seg = "FUVA"
    seg2 = "NUVB"
    # Test
    key = cosutil.segmentSpecificKeyword(root, seg)  # keyword_root_a
    key2 = cosutil.segmentSpecificKeyword(root, seg2)  # keyword_root_
    # Verify
    root2 = root
    root += "a"
    assert key == root
    assert key2 == root2


def test_find_ref_file(tmp_path):
    filename = str(tmp_path / "test.fits")
    wrong_filename = str(tmp_path / "wrong_file.fits")
    generate_fits_file(filename)
    generate_fits_file(wrong_filename)
    fits.setval(wrong_filename, 'FILETYPE', value='FLAT FIELD REFERENCE IMAGE')
    # Missing
    ref1 = {"keyword": "FLATFILE", "filename": "test_flt.fits", "calcos_ver": "3.0",
            "min_ver": "2.2", "filetype": "FLAT FIELD REFERENCE IMAGE"}
    # Bad version
    ref2 = {"keyword": "FLATFILE", "filename": filename, "calcos_ver": "2.21",
            "min_ver": "3.3", "filetype": "FLAT FIELD REFERENCE IMAGE"}
    # Wrong file
    ref3 = {"keyword": "FLATFILE", "filename": wrong_filename, "calcos_ver": "3.0",
            "min_ver": "2.3", "filetype": "IMAGE"}
    missing1 = {}
    missing2 = {}
    missing3 = {}
    wrong_f1 = {}
    wrong_f2 = {}
    wrong_f3 = {}
    bad_ver1 = {}
    bad_ver2 = {}
    bad_ver3 = {}
    # Actual values
    actual_missing = {"FLATFILE": "test_flt.fits"}
    actual_bad_ver = {"FLATFILE": (filename, "  the reference file must be at least version 3.3")}
    actual_wrong_ver = {'FLATFILE': (wrong_filename, 'IMAGE')}
    # Test
    cosutil.findRefFile(ref1, missing1, wrong_f1, bad_ver1)
    cosutil.findRefFile(ref2, missing2, wrong_f2, bad_ver2)
    cosutil.findRefFile(ref3, missing3, wrong_f3, bad_ver3)
    # Verify
    assert actual_missing == missing1
    assert actual_bad_ver == bad_ver2
    assert actual_wrong_ver == wrong_f3


def test_cmp_version():
    # Setup
    min_ver = ["1", "1", "1.1", "1.1", "1.1", "1.2", "1.0", "2.7", "2.0", "2.9", "2.12d", "2.13d", "2.13"]
    vcalcos = ["1", "1.1", "1", "1.1", "1.2", "1.1", "1.7", "2.8", "2.13.1", "2.9", "2.13b", "2.13b", "2.13b"]
    calcos_ver = ["1.1", "1", "1", "1.2", "1.1", "1.1", "2.3", "2.8a", "2.13", "2.13.1", "2.12a", "2.13a", "2.13c"]

    expected_cmp = [0, 1, -1, 0, 1, -1, 0, 0, 1, 0, 1, -1, 0]
    # Test and Verify
    test_cmp = []
    for i in range(len(expected_cmp)):
        test_cmp.append(cosutil.cmpVersion(min_ver[i], vcalcos[i], calcos_ver[i]))
        assert expected_cmp[i] == test_cmp[i]


def test_get_pedigree(tmp_path):
    # Setup
    capture_msg = io.StringIO()
    sys.stdout = capture_msg
    switch = "perform"
    refkey = "statflag"
    filename = str(tmp_path / "test_flt.file")
    generate_fits_file(filename)
    err_msg = f"Warning:  STATFLAG {filename} is a dummy file\n" \
              "    so PERFORM will not be done.\n"
    # Test
    pedgr1 = cosutil.getPedigree(switch, refkey, filename)
    fits.setval(filename, "pedigree", value="DUMMY", ext=0)
    pedgr2 = cosutil.getPedigree(switch, refkey, filename)
    sys.stdout = sys.__stdout__
    print(capture_msg.getvalue())

    # Verify
    assert pedgr1 == "OK"
    assert pedgr2 == "DUMMY"
    assert err_msg == capture_msg.getvalue()


def test_get_aperture_keyword(tmp_path):
    filename = str(tmp_path / "aperture_test.fits")
    # Setup
    generate_fits_file(filename)
    # condition 1
    fits.setval(filename, "aperture", value="PSA-FUV", ext=0)
    fits.setval(filename, "propaper", value="PSA-FUV", ext=0)
    hdr1 = fits.getheader(filename, ext=0)
    # condition 2
    fits.setval(filename, "aperture", value="RelMvReq", ext=0)
    fits.setval(filename, "propaper", value="WCA", ext=0)
    hdr2 = fits.getheader(filename, ext=0)
    # condition 3
    fits.setval(filename, "propaper", value="NA", ext=0)
    fits.setval(filename, "shutter", value="closed", ext=0)
    fits.setval(filename, "lampused", value="P", ext=0)
    hdr3 = fits.getheader(filename, ext=0)
    # condition 4
    fits.setval(filename, "lampused", value="D", ext=0)
    hdr4 = fits.getheader(filename, ext=0)
    # condition 5
    fits.setval(filename, "lampused", value="A", ext=0)
    hdr5 = fits.getheader(filename, ext=0)
    # condition 6
    fits.setval(filename, "shutter", value="open", ext=0)
    fits.setval(filename, "life_adj", value=2, ext=0)
    fits.setval(filename, "aperypos", value=4.56, ext=0)
    hdr6 = fits.getheader(filename, ext=0)

    # Expected values
    rtn1 = ('PSA', 'APERTURE changed from PSA-FUV to PSA')
    rtn2 = ('WCA', 'APERTURE changed from RelMvReq to WCA (copied from PROPAPER)')
    rtn3 = ('WCA', 'Guessing correct APERTURE ... was RelMvReq, now set to WCA')
    rtn4 = ('FCA', 'Guessing correct APERTURE ... was RelMvReq, now set to FCA')
    rtn5 = ('PSA', 'Guessing correct APERTURE ... was RelMvReq, now set to PSA')
    rtn6 = (None, 'Guessing correct APERTURE ... was RelMvReq, now set to PSA')
    # Test
    rtn_value1 = cosutil.getApertureKeyword(hdr1)
    rtn_value2 = cosutil.getApertureKeyword(hdr2)
    rtn_value3 = cosutil.getApertureKeyword(hdr3)
    rtn_value4 = cosutil.getApertureKeyword(hdr4)
    rtn_value5 = cosutil.getApertureKeyword(hdr5)
    rtn_value6 = cosutil.getApertureKeyword(hdr6)
    # Verify
    assert rtn1 == rtn_value1
    assert rtn2 == rtn_value2
    assert rtn3 == rtn_value3
    assert rtn4 == rtn_value4
    assert rtn5 == rtn_value5
    assert rtn6 == rtn_value6


def test_write_version_to_trailer(tmp_path):
    capture_msg = io.StringIO()
    sys.stdout = capture_msg
    generate_fits_file(str(tmp_path / "dummy_file.fits"))
    ascii_file = open(str(tmp_path / "ascii.txt"), mode="w")
    cosutil.fd_trl = ascii_file
    cosutil.CALCOS_VERSION = '3.1.0'
    cosutil.writeVersionToTrailer()
    sys.stdout = sys.__stdout__
    assert ascii_file == cosutil.fd_trl
    assert capture_msg.getvalue() == ''

def test_get_switch(tmp_path):
    # Setup
    filename = str(tmp_path / "switch.fits")
    generate_fits_file(filename)
    phdr = fits.getheader(filename, ext=0)
    keyword = "statflag"
    # Test
    switch = cosutil.getSwitch(phdr, keyword)
    phdr.set('STATFLAG', False, 'Calculate statistics')
    switch2 = cosutil.getSwitch(phdr, keyword)
    switch3 = cosutil.getSwitch(phdr, "none")
    # Verify
    assert switch == 'PERFORM'
    assert switch2 == 'OMIT'
    assert switch3 == 'N/A'


def test_temp_pulse_height_range(tmp_path):
    filename = str(tmp_path / "pulseHeightRef.fits")
    generate_fits_file(filename)
    true_pha_value = 4
    fits.setval(filename, "pharange", value=true_pha_value, ext=0)
    fits.getheader(filename, ext=0)
    pha_value = cosutil.tempPulseHeightRange(filename)
    assert true_pha_value == pha_value


def test_get_pulse_height_range(tmp_path):
    filename = str(tmp_path / "pulseHeightRef.fits")
    generate_fits_file(filename)
    fits.setval(filename, "phalowrA", value=7, ext=0)
    fits.setval(filename, "phaupprA", value=10, ext=0)
    hdu = fits.getheader(filename, ext=0)
    seg = ['FUVA', 'FUVB']
    actual = [' 7_10', None]
    for s, a in zip(seg, actual):
        test_str = cosutil.getPulseHeightRange(hdu, s)
        assert a == test_str


def test_time_at_midpoint(tmp_path):
    # Setup
    filename = str(tmp_path / "test_timeAtMidpoint.fits")
    generate_fits_file(filename)
    hdr = fits.getheader(filename, ext=1)
    info = {'expstart': hdr['TCRPX7'], 'expend': hdr['TCRVL7']}
    average = 4776.314586556611
    # Test
    test_average = cosutil.timeAtMidpoint(info)
    # Verify
    assert average == test_average


def test_timeline_times():
    # Setup
    first_time = np.array([2.5], dtype=np.float32)
    last_time = np.array([3.8], dtype=np.float32)
    actual_values = [2.5, 3.5, 4.5, 5.5, 6.5, 7.5, 8.5, 9.5, 10.5]
    # Test
    tl_time = cosutil.timelineTimes(2.5, 10.8)
    # Verify
    for i in range(len(actual_values)):
        assert actual_values[i] == tl_time[i]


def test_combine_stat():
    # Setup
    stat_info = [{'ngoodpix': 5.8, "sci_goodmax": 2.7, "sci_goodmean": 4.3, "err_goodmax": 1,
                  "err_goodmean": 2}, {'ngoodpix': 10.8, "sci_goodmax": 4.7, "sci_goodmean": 8.3, "err_goodmax": 2,
                                       "err_goodmean": 1}]
    actual = {'ngoodpix': 16.6, 'sci_goodmax': 4.7, 'sci_goodmean': 6.902409638554217, 'err_goodmax': 2,
              'err_goodmean': 1.3493975903614457}
    # Test
    test = cosutil.combineStat(stat_info)
    # Verify
    for key in actual.keys():
        assert actual[key] == test[key]


def test_override_keywords(tmp_path):
    # Setup
    filename = str(tmp_path / "overridekeywords.fits")
    generate_fits_file(filename)
    info = {"cal_ver": 3.1, "opt_elem": 2, "cenwave": 0.34, "fpoffset": 3.43, "obstype": "FUV",
            "exptype": "N/A", "aperture": "PSA", "x_offset": 1.2, "dispaxis": 2.5}
    switches = {"statflag": "PERFORM", "flatcorr": "PERFORM", "geocorr": "COMPLETE", "randcorr": "SKIPPED"}
    reffiles = {"flatfile": "abc_flat.fits", "flt_hdr": "lref$abc_flat.fits"}
    fits.setval(filename, "flt_hdr", value="NA", ext=0)
    fits.setval(filename, "dispaxis", value=2.9, ext=1)
    fits.setval(filename, "x_offset", value=1.9, ext=1)
    phdr = fits.getheader(filename, ext=0)
    hdr = fits.getheader(filename, ext=1)
    # Actual Values
    val1 = True
    val2 = switches.values()
    val3 = phdr.values()
    val4 = reffiles["flt_hdr"]
    # Test
    cosutil.overrideKeywords(phdr, hdr, info, switches, reffiles)
    phdr = fits.getheader(filename, ext=0)
    # Verify
    assert val1 == phdr["statflag"]



import os

import numpy as np
from astropy.io import fits

from calcos import average
from generate_tempfiles import create_count_file


def test_avg_image(tmp_path):
    """
    tests avg_image() in average.py
    explanation of the test
    - create temporary count files to be used as inputs
    - expected values in the output file are the average of the input values
    - loop though the values to check if the math holds.
    Returns
    -------
    pass if expected == actual fail otherwise.
    """
    # Setup
    infile = [str(tmp_path / "test_count1.fits"), str(tmp_path / "test_count2.fits")]
    outfile = str(tmp_path / "test_output.fits")
    create_count_file(infile[0])
    create_count_file(infile[1])
    inhdr1, inhdr2 = fits.open(infile[0]), fits.open(infile[1])
    # Test
    average.avgImage(infile, outfile)
    out_hdr = fits.open(outfile)

    # Verify
    assert os.path.exists(outfile)
    for (i, j, k) in zip(inhdr1[1].header, inhdr2[1].header, out_hdr[1].header):
        assert i == j == k
    np.testing.assert_array_equal((inhdr1[1].data + inhdr1[1].data) / 2, out_hdr[1].data)
